# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import ultipa.proto.ultipa_pb2 as ultipa__pb2

class UltipaRpcsStub(object):
    """The ultipa service definition.
    """

    def __init__(self, channel):
        """
        Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SayHello = channel.unary_unary(
                '/ultipa.UltipaRpcs/SayHello',
                request_serializer=ultipa__pb2.HelloUltipaRequest.SerializeToString,
                response_deserializer=ultipa__pb2.HelloUltipaReply.FromString,
                )
        self.Uql = channel.unary_stream(
                '/ultipa.UltipaRpcs/Uql',
                request_serializer=ultipa__pb2.UqlRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UqlReply.FromString,
                )
        self.InsertNodes = channel.unary_unary(
                '/ultipa.UltipaRpcs/InsertNodes',
                request_serializer=ultipa__pb2.InsertNodesRequest.SerializeToString,
                response_deserializer=ultipa__pb2.InsertNodesReply.FromString,
                )
        self.InsertEdges = channel.unary_unary(
                '/ultipa.UltipaRpcs/InsertEdges',
                request_serializer=ultipa__pb2.InsertEdgesRequest.SerializeToString,
                response_deserializer=ultipa__pb2.InsertEdgesReply.FromString,
                )


class UltipaRpcsServicer(object):
    """The ultipa service definition.
    """

    def SayHello(self, request, context):
        """1. Send a greeting
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Uql(self, request, context):
        """2. uql
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InsertNodes(self, request, context):
        """3. Insert nodes
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InsertEdges(self, request, context):
        """4. Insert edges
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UltipaRpcsServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SayHello': grpc.unary_unary_rpc_method_handler(
                    servicer.SayHello,
                    request_deserializer=ultipa__pb2.HelloUltipaRequest.FromString,
                    response_serializer=ultipa__pb2.HelloUltipaReply.SerializeToString,
            ),
            'Uql': grpc.unary_stream_rpc_method_handler(
                    servicer.Uql,
                    request_deserializer=ultipa__pb2.UqlRequest.FromString,
                    response_serializer=ultipa__pb2.UqlReply.SerializeToString,
            ),
            'InsertNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.InsertNodes,
                    request_deserializer=ultipa__pb2.InsertNodesRequest.FromString,
                    response_serializer=ultipa__pb2.InsertNodesReply.SerializeToString,
            ),
            'InsertEdges': grpc.unary_unary_rpc_method_handler(
                    servicer.InsertEdges,
                    request_deserializer=ultipa__pb2.InsertEdgesRequest.FromString,
                    response_serializer=ultipa__pb2.InsertEdgesReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ultipa.UltipaRpcs', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UltipaRpcs(object):
    """The ultipa service definition.
    """

    @staticmethod
    def SayHello(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaRpcs/SayHello',
            ultipa__pb2.HelloUltipaRequest.SerializeToString,
            ultipa__pb2.HelloUltipaReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Uql(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/ultipa.UltipaRpcs/Uql',
            ultipa__pb2.UqlRequest.SerializeToString,
            ultipa__pb2.UqlReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InsertNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaRpcs/InsertNodes',
            ultipa__pb2.InsertNodesRequest.SerializeToString,
            ultipa__pb2.InsertNodesReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InsertEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaRpcs/InsertEdges',
            ultipa__pb2.InsertEdgesRequest.SerializeToString,
            ultipa__pb2.InsertEdgesReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class UltipaControlsStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """
        Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SayHello = channel.unary_unary(
                '/ultipa.UltipaControls/SayHello',
                request_serializer=ultipa__pb2.HelloUltipaRequest.SerializeToString,
                response_deserializer=ultipa__pb2.HelloUltipaReply.FromString,
                )
        self.UserSetting = channel.unary_unary(
                '/ultipa.UltipaControls/UserSetting',
                request_serializer=ultipa__pb2.UserSettingRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UserSettingReply.FromString,
                )
        self.GetLeader = channel.unary_unary(
                '/ultipa.UltipaControls/GetLeader',
                request_serializer=ultipa__pb2.GetLeaderRequest.SerializeToString,
                response_deserializer=ultipa__pb2.GetLeaderReply.FromString,
                )
        self.UqlEx = channel.unary_stream(
                '/ultipa.UltipaControls/UqlEx',
                request_serializer=ultipa__pb2.UqlRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UqlReply.FromString,
                )
        self.Export = channel.unary_stream(
                '/ultipa.UltipaControls/Export',
                request_serializer=ultipa__pb2.ExportRequest.SerializeToString,
                response_deserializer=ultipa__pb2.ExportReply.FromString,
                )
        self.DownloadFileV2 = channel.unary_stream(
                '/ultipa.UltipaControls/DownloadFileV2',
                request_serializer=ultipa__pb2.DownloadFileRequestV2.SerializeToString,
                response_deserializer=ultipa__pb2.DownloadFileReply.FromString,
                )
        self.InstallAlgo = channel.stream_unary(
                '/ultipa.UltipaControls/InstallAlgo',
                request_serializer=ultipa__pb2.InstallAlgoRequest.SerializeToString,
                response_deserializer=ultipa__pb2.InstallAlgoReply.FromString,
                )
        self.UninstallAlgo = channel.unary_unary(
                '/ultipa.UltipaControls/UninstallAlgo',
                request_serializer=ultipa__pb2.UninstallAlgoRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UninstallAlgoReply.FromString,
                )
        self.Uploader = channel.stream_unary(
                '/ultipa.UltipaControls/Uploader',
                request_serializer=ultipa__pb2.UploaderRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UploaderReply.FromString,
                )
        self.CreateGraphByUploader = channel.unary_unary(
                '/ultipa.UltipaControls/CreateGraphByUploader',
                request_serializer=ultipa__pb2.CreateGraphByUploaderRequest.SerializeToString,
                response_deserializer=ultipa__pb2.CreateGraphByUploaderReply.FromString,
                )
        self.InstallExta = channel.stream_unary(
                '/ultipa.UltipaControls/InstallExta',
                request_serializer=ultipa__pb2.InstallExtaRequest.SerializeToString,
                response_deserializer=ultipa__pb2.InstallExtaReply.FromString,
                )
        self.UninstallExta = channel.unary_unary(
                '/ultipa.UltipaControls/UninstallExta',
                request_serializer=ultipa__pb2.UninstallExtaRequest.SerializeToString,
                response_deserializer=ultipa__pb2.UninstallExtaReply.FromString,
                )
        self.Authenticate = channel.unary_unary(
                '/ultipa.UltipaControls/Authenticate',
                request_serializer=ultipa__pb2.AuthenticateRequest.SerializeToString,
                response_deserializer=ultipa__pb2.AuthenticateReply.FromString,
                )


class UltipaControlsServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SayHello(self, request, context):
        """1. Send a greeting
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UserSetting(self, request, context):
        """2. User settings that stores user configuration, controlled by user
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLeader(self, request, context):
        """3. Get the raft leader
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UqlEx(self, request, context):
        """4. Uql extension, the interface used for the execution of top(), kill(), show().* and stats()
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Export(self, request, context):
        """5. Export nodes and edges
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadFileV2(self, request, context):
        """6. Download a file generated by the writeback operation of an algorithm
        The 2nd version, needs to input arguments 'algo name' and 'task ID'
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InstallAlgo(self, request_iterator, context):
        """7. Install an Ultipa standard algorithm
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UninstallAlgo(self, request, context):
        """8. Uninstall an Ultipa standard algorithm
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Uploader(self, request_iterator, context):
        """9. Remote graph loader
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateGraphByUploader(self, request, context):
        """10. Create graph by uploader
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InstallExta(self, request_iterator, context):
        """11. Install an EXTA algorithm
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UninstallExta(self, request, context):
        """12. Uninstall an EXTA algorithm
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Authenticate(self, request, context):
        """13. Authenticate a user
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UltipaControlsServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SayHello': grpc.unary_unary_rpc_method_handler(
                    servicer.SayHello,
                    request_deserializer=ultipa__pb2.HelloUltipaRequest.FromString,
                    response_serializer=ultipa__pb2.HelloUltipaReply.SerializeToString,
            ),
            'UserSetting': grpc.unary_unary_rpc_method_handler(
                    servicer.UserSetting,
                    request_deserializer=ultipa__pb2.UserSettingRequest.FromString,
                    response_serializer=ultipa__pb2.UserSettingReply.SerializeToString,
            ),
            'GetLeader': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLeader,
                    request_deserializer=ultipa__pb2.GetLeaderRequest.FromString,
                    response_serializer=ultipa__pb2.GetLeaderReply.SerializeToString,
            ),
            'UqlEx': grpc.unary_stream_rpc_method_handler(
                    servicer.UqlEx,
                    request_deserializer=ultipa__pb2.UqlRequest.FromString,
                    response_serializer=ultipa__pb2.UqlReply.SerializeToString,
            ),
            'Export': grpc.unary_stream_rpc_method_handler(
                    servicer.Export,
                    request_deserializer=ultipa__pb2.ExportRequest.FromString,
                    response_serializer=ultipa__pb2.ExportReply.SerializeToString,
            ),
            'DownloadFileV2': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadFileV2,
                    request_deserializer=ultipa__pb2.DownloadFileRequestV2.FromString,
                    response_serializer=ultipa__pb2.DownloadFileReply.SerializeToString,
            ),
            'InstallAlgo': grpc.stream_unary_rpc_method_handler(
                    servicer.InstallAlgo,
                    request_deserializer=ultipa__pb2.InstallAlgoRequest.FromString,
                    response_serializer=ultipa__pb2.InstallAlgoReply.SerializeToString,
            ),
            'UninstallAlgo': grpc.unary_unary_rpc_method_handler(
                    servicer.UninstallAlgo,
                    request_deserializer=ultipa__pb2.UninstallAlgoRequest.FromString,
                    response_serializer=ultipa__pb2.UninstallAlgoReply.SerializeToString,
            ),
            'Uploader': grpc.stream_unary_rpc_method_handler(
                    servicer.Uploader,
                    request_deserializer=ultipa__pb2.UploaderRequest.FromString,
                    response_serializer=ultipa__pb2.UploaderReply.SerializeToString,
            ),
            'CreateGraphByUploader': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateGraphByUploader,
                    request_deserializer=ultipa__pb2.CreateGraphByUploaderRequest.FromString,
                    response_serializer=ultipa__pb2.CreateGraphByUploaderReply.SerializeToString,
            ),
            'InstallExta': grpc.stream_unary_rpc_method_handler(
                    servicer.InstallExta,
                    request_deserializer=ultipa__pb2.InstallExtaRequest.FromString,
                    response_serializer=ultipa__pb2.InstallExtaReply.SerializeToString,
            ),
            'UninstallExta': grpc.unary_unary_rpc_method_handler(
                    servicer.UninstallExta,
                    request_deserializer=ultipa__pb2.UninstallExtaRequest.FromString,
                    response_serializer=ultipa__pb2.UninstallExtaReply.SerializeToString,
            ),
            'Authenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.Authenticate,
                    request_deserializer=ultipa__pb2.AuthenticateRequest.FromString,
                    response_serializer=ultipa__pb2.AuthenticateReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ultipa.UltipaControls', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UltipaControls(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SayHello(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/SayHello',
            ultipa__pb2.HelloUltipaRequest.SerializeToString,
            ultipa__pb2.HelloUltipaReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UserSetting(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/UserSetting',
            ultipa__pb2.UserSettingRequest.SerializeToString,
            ultipa__pb2.UserSettingReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetLeader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/GetLeader',
            ultipa__pb2.GetLeaderRequest.SerializeToString,
            ultipa__pb2.GetLeaderReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UqlEx(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/ultipa.UltipaControls/UqlEx',
            ultipa__pb2.UqlRequest.SerializeToString,
            ultipa__pb2.UqlReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Export(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/ultipa.UltipaControls/Export',
            ultipa__pb2.ExportRequest.SerializeToString,
            ultipa__pb2.ExportReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DownloadFileV2(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/ultipa.UltipaControls/DownloadFileV2',
            ultipa__pb2.DownloadFileRequestV2.SerializeToString,
            ultipa__pb2.DownloadFileReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InstallAlgo(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/ultipa.UltipaControls/InstallAlgo',
            ultipa__pb2.InstallAlgoRequest.SerializeToString,
            ultipa__pb2.InstallAlgoReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UninstallAlgo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/UninstallAlgo',
            ultipa__pb2.UninstallAlgoRequest.SerializeToString,
            ultipa__pb2.UninstallAlgoReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Uploader(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/ultipa.UltipaControls/Uploader',
            ultipa__pb2.UploaderRequest.SerializeToString,
            ultipa__pb2.UploaderReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateGraphByUploader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/CreateGraphByUploader',
            ultipa__pb2.CreateGraphByUploaderRequest.SerializeToString,
            ultipa__pb2.CreateGraphByUploaderReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def InstallExta(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/ultipa.UltipaControls/InstallExta',
            ultipa__pb2.InstallExtaRequest.SerializeToString,
            ultipa__pb2.InstallExtaReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UninstallExta(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/UninstallExta',
            ultipa__pb2.UninstallExtaRequest.SerializeToString,
            ultipa__pb2.UninstallExtaReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Authenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ultipa.UltipaControls/Authenticate',
            ultipa__pb2.AuthenticateRequest.SerializeToString,
            ultipa__pb2.AuthenticateReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:36
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Algo.py
from ultipa.structs.BaseModel import BaseModel


class Algo(BaseModel):
	'''
	    Data class for algorithm.
	'''
	name: str
	description: str
	version: str
	result_opt: str
	parameters: dict
	write_to_stats_parameters: dict
	write_to_db_parameters: dict
	write_to_file_parameters: dict
	# write_to_client_normal_parameters:dict
	# write_to_client_stream_parameters:dict

	def __init__(self,name:str,description:str,version:str,result_opt:str,parameters:dict,
			  write_to_stats_parameters:dict=None,write_to_db_parameters:dict=None,
			  write_to_file_parameters:dict=None
			#   write_to_client_normal_parameters:dict=None,
			#   write_to_client_stream_parameters:dict=None
			  ):
		self.name = name
		self.description = description
		self.version = version
		self.result_opt = result_opt
		self.parameters = parameters
		self.write_to_stats_parameters = write_to_stats_parameters
		self.write_to_db_parameters=write_to_db_parameters
		self.write_to_file_parameters=write_to_file_parameters
		# self.write_to_client_normal_parameters=write_to_client_normal_parameters
		# self.write_to_client_stream_parameters=write_to_client_stream_parameters


class AlgoResultOpt(BaseModel):
	can_realtime: bool
	can_visualization: bool
	can_write_back: bool

class ALGO_RESULT:
	ALGO_RESULT_UNSET = -1
	WRITE_TO_FILE = 1
	WRITE_TO_DB = 2
	WRITE_TO_CLIENT = 4
	WRITE_TO_VISUALIZATION = 8


from typing import List
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType
from ultipa.configuration.InsertConfig import InsertConfig
from ultipa.configuration.InsertRequestConfig import InsertRequestConfig
from ultipa.structs.Node import Node
from ultipa.utils.ufilter.new_ufilter import *

class NodeExtra(BaseExtra):
	'''
	Processing class that defines settings for node related operations.

	'''
	def searchNode(self, request: ULTIPA_REQUEST.SearchNode,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseSearchNode:
		'''
		Query for nodes.

		Args:
			request: An object of SearchNode class

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseSearchNode

		'''

		uqlMaker = UQLMAKER(command=CommandList.nodes, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		uqlMaker.addParam('as', request.select.aliasName)
		uqlMaker.addParam("return", request.select)
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res

		return res

	def insertNodes(self, nodes: List[Node], schemaName: str, config:InsertRequestConfig) -> ULTIPA_RESPONSE.ResponseInsertNode:
		'''
		Insert nodes.

		Args:
			nodes: List of nodes to be inserted

			schemaName: name of the Schema

			config: An object of InsertConfig classd

		Returns:
			ResponseInsertNode
		'''

		combined_values = []  # to combine values and id for insertion
		for node in nodes:
			node_dict = {}
			if node.id:
				node_dict["_id"] = node.id
			if node.uuid:
				node_dict["_uuid"]= node.uuid	
			node_dict.update(node.values)
			combined_values.append(node_dict)

		nodes=combined_values
		schemaName='@' + schemaName

		uqlMaker = UQLMAKER(command=CommandList.insert, commonParams=config)
		if config.insertType==InsertType.UPSERT:
			uqlMaker = UQLMAKER(command=CommandList.upsert, commonParams=config)
		if config.insertType==InsertType.OVERWRITE:
			uqlMaker.addParam('overwrite', "", required=False)
		if schemaName:
			uqlMaker.addParam('into', schemaName, required=False)

		uqlMaker.addParam('nodes', nodes)

		if config.silent==False:
			uqlMaker.addParam('as', "nodes")
			# uqlMaker.addParam('return', "nodes._uuid")
			uqlMaker.addParam('return', "nodes{*}")

		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		if config.silent==False:
			if len(res.aliases) > 0:
				res.data = res.items.get(res.aliases[0].alias).data
		return res

	def updateNodes(self, request: ULTIPA_REQUEST.UpdateNode,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Update nodes.

		Args:
			request: An object of UpdateNode class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		uqlMaker = UQLMAKER(command=CommandList.updateNodes, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		uqlMaker.addParam("set", request.values)
		uqlMaker.addParam("silent", request.silent)
		res = self.uqlSingle(uqlMaker)
		return res

	def deleteNodes(self,filter:str, config:InsertRequestConfig) -> ULTIPA_RESPONSE.ResponseDeleteNode:
		'''
		Delete nodes.

		Args:
			filter: An object of UltipaFilter class

			config: An object of InsertConfig class

		Returns:
			ResponseDeleteNode
		'''

		uqlMaker = UQLMAKER(command=CommandList.deleteNodes, commonParams=config)
		# if request.id:
		# 	uqlMaker.setCommandParams(request.id)
		if filter:
			uqlMaker.setCommandParams(filter)

		if config.silent==False:
			uqlMaker.addParam('as', "nodes")
			# uqlMaker.addParam('return', "nodes._uuid")	
			uqlMaker.addParam('return', "nodes{*}")
		res = self.uqlSingle(uqlMaker)

		if config.silent==False:
			if len(res.aliases) > 0:
				res.data = res.items.get(res.aliases[0].alias).data
		return res




class SystemPrivileges:
    '''
        Data class for system privileges.
    '''

    TRUNCATE = 'TRUNCATE'
    COMPACT = 'COMPACT'
    CREATE_GRAPH = 'CREATE_GRAPH'
    SHOW_GRAPH = 'SHOW_GRAPH'
    DROP_GRAPH = 'DROP_GRAPH'
    ALTER_GRAPH = 'ALTER_GRAPH'
    MOUNT_GRAPH = 'MOUNT_GRAPH'
    UNMOUNT_GRAPH = 'UNMOUNT_GRAPH'
    TOP = 'TOP'
    KILL = 'KILL'
    STAT = 'STAT'
    SHOW_POLICY = 'SHOW_POLICY'
    CREATE_POLICY = 'CREATE_POLICY'
    DROP_POLICY = 'DROP_POLICY'
    ALTER_POLICY = 'ALTER_POLICY'
    SHOW_USER = 'SHOW_USER'
    CREATE_USER = 'CREATE_USER'
    DROP_USER = 'DROP_USER'
    ALTER_USER = 'ALTER_USER'
    SHOW_PRIVILEGE = 'SHOW_PRIVILEGE'


class GraphPrivileges:
    '''
        Data class for graphset privileges.
    '''
    TEMPLATE = "TEMPLATE"
    KHOP = "KHOP"
    AB = "AB"
    SPREAD = "SPREAD"
    AUTONET = "AUTONET"
    FIND = "FIND"
    INSERT = "INSERT"
    EXPORT = "EXPORT"
    UPSERT = "UPSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    CREATE_SCHEMA = "CREATE_SCHEMA"
    DROP_SCHEMA = "DROP_SCHEMA"
    ALTER_SCHEMA = "ALTER_SCHEMA"
    SHOW_SCHEMA = "SHOW_SCHEMA"
    CREATE_PROPERTY = "CREATE_PROPERTY"
    DROP_PROPERTY = "DROP_PROPERTY"
    ALTER_PROPERTY = "ALTER_PROPERTY"
    SHOW_PROPERTY = "SHOW_PROPERTY"
    CREATE_FULLTEXT = "CREATE_FULLTEXT"
    DROP_FULLTEXT = "DROP_FULLTEXT"
    SHOW_FULLTEXT = "SHOW_FULLTEXT"
    CREATE_INDEX = "CREATE_INDEX"
    DROP_INDEX = "DROP_INDEX"
    SHOW_INDEX = "SHOW_INDEX"
    LTE = "LTE"
    UFE = "UFE"
    CLEAR_TASK = "CLEAR_TASK"
    STOP_TASK = "STOP_TASK"
    PAUSE_TASK = "PAUSE_TASK"
    RESUME_TASK = "RESUME_TASK"
    SHOW_TASK = "SHOW_TASK"
    ALGO = "ALGO"
    SHOW_ALGO = "SHOW_ALGO"


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:46
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : requestConfig.py
from ultipa.structs.Retry import Retry


class RequestConfig:
	'''
    Configuration class for the interface of any non-insert operation.

    This class provides settings for all the operations other than inserting metadata.
    
    Args:
        - graphName (str): The name of graphset to use.
        - timeout (int): The timeout for the request in seconds.
        - retry (Retry): The object of Retry configuration class when the request fails
        - stream (bool): Whether to return stream
        - host (str): The designated host the request will be sent to, or sent to a random host if not set
        - useMaster (bool):	Whether to send the request to the leader to guarantee Consistency Read
        - threadNum (int): The number of threads used for the request
        - timeZone (str): The string of timezone in standard format
        - timeZoneOffset (any): 1, the number of seconds; 2, a 5-character string such as +0700, -0430
	'''

	def __init__(self, graphName: str = '', timeout: int = 3600, retry: Retry = Retry(),
				 stream: bool = False, host: str = None, useMaster: bool = False, threadNum: int = None,
				 timeZone: str = None, timeZoneOffset: any = None):
		self.graphName = graphName
		self.timeoutWithSeconds = timeout
		self.retry = retry
		self.stream = stream
		self.useHost = host
		self.useMaster = useMaster
		self.threadNum = threadNum
		self.timeZone = timeZone
		self.timeZoneOffset = timeZoneOffset




from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.structs import Schema
from ultipa.structs import DBType
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from typing import Tuple
from ultipa.operate.property_extra import PropertyExtra
from ultipa.structs.Property import Property

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
	"name": "schemaName",
	"type": "propertyType",
}


class SchemaExtra(BaseExtra):
	'''
		Prcessing class that defines settings for schema related operations.
	'''

	def createSchema(self, schema: Schema, isCreateProperties:bool = False,
					 requestConfig: RequestConfig = RequestConfig())->ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a schema.

		Args:
			schema: An object of Schema class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		command = schema.DBType == DBType.DBNODE and CommandList.createNodeSchema or CommandList.createEdgeSchema
		commandP = [f"`{schema.name}`"]
		if schema.description:
			commandP.append(schema.description)
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
	
		if isCreateProperties:
			if schema.properties:
				for prop in schema.properties:
					try:
						res1 = PropertyExtra.createProperty(self, dbType=schema.DBType, schemaName=schema.name, prop=prop)
					except Exception as e:
						print("An error occurred while creating property:", prop.name, "Error:", e)		
		return res


	def showSchema(self,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListSchema:
		'''
		Show schema(s).

		Args:			
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListSchema

		'''

		command = CommandList.showSchema		
		commandP = ''

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker)
		schemalist=[]
		nodeschemas=res.alias("_nodeSchema").asSchemas()
		edgeschemas=res.alias("_edgeSchema").asSchemas()
		schemalist.extend(nodeschemas)
		schemalist.extend(edgeschemas)
		res.items=schemalist
		return res
	


	def showNodeSchema(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListSchema:
		'''
		Show Nodeschema(s).

		Args:			
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListSchema

		'''

		command=CommandList.showNodeSchema
		commandp=''
		# 	if schemaName:
		# 		commandP = '@' + schemaName
		# 	else:
		# 		commandP = ''
		# else:
		# 	command = CommandList.showSchema
		# 	commandP = ''
		uqlMaker = UQLMAKER(command=command,commandP=commandp, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker)
		res.items = res.alias("_nodeSchema").asSchemas()
		return res
	
	def showEdgeSchema(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListSchema:
		'''
		Show Edgeschema(s).

		Args:			
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListSchema

		'''

		command=CommandList.showEdgeSchema
		commandp=''
		# 	if schemaName:
		# 		commandP = '@' + schemaName
		# 	else:
		# 		commandP = ''
		# else:
		# 	command = CommandList.showSchema
		# 	commandP = ''

		# uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		uqlMaker = UQLMAKER(command=command,commandP=commandp, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker)
		res.items = res.alias("_edgeSchema").asSchemas()
		return res


	def alterSchema(self, schema:Schema,newSchema:Schema,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Alter schema.

		Args:
			schema: An object of schema class(to be altered)
			newSchema: An object of schema class

		Returns:
			 UltipaResponse

		'''
		command = schema.DBType == DBType.DBNODE and CommandList.alterNodeSchema or CommandList.alterEdgeSchema
		# commandP = '@' + schema.name
		commandP="@`%s`" % (schema.name)
		update_dict = {}
		if newSchema.name:
			update_dict.setdefault('name', newSchema.name)
		if newSchema.description:
			update_dict.update({'description': newSchema.description})
		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		uqlMaker.addParam("set", update_dict)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropSchema(self, schema:Schema,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop schema.

		Args:
			schema: An object of schema class

		Returns:
			 UltipaResponse

		'''

		command = schema.DBType == DBType.DBNODE and CommandList.dropNodeSchema or CommandList.dropEdgeSchema
		# commandP = '@' + schema.name
		commandP="@`%s`" % (schema.name)

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res
	
	def getSchema(self, schemaName: str, dbType: DBType,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseSchema:
		'''
		Acquire a designated Schema.

		Args:
			schemaName: The name of Schema
			
			dbType:The DBType of data (DBNODE or DBEDGE)

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseSchema
		'''

		
		if dbType != None:
			if dbType == DBType.DBNODE:
				command = CommandList.showNodeSchema
			elif dbType == DBType.DBEDGE:
				command = CommandList.showEdgeSchema

			if schemaName:
				commandP="@`%s`" % (schemaName)


		uqlMaker = UQLMAKER(command, commandP=commandP,commonParams=requestConfig)

		res = self.uqlSingle(uqlMaker=uqlMaker)
		if res.status.code==ULTIPA.Code.SUCCESS:
			if dbType==DBType.DBNODE:
				res.items = res.alias("_nodeSchema").asSchemas()[0]
			else:
				res.items = res.alias("_edgeSchema").asSchemas()[0]
		return res	
	

	def getNodeSchema(self, schemaName: str,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseSchema:
		'''
		Acquire a designated Node Schema.

		Args:
			schemaName: The name of Schema
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseSchema
		'''

		
		command=CommandList.showNodeSchema

		if schemaName:
			commandP="@`%s`" % (schemaName)


		uqlMaker = UQLMAKER(command, commandP=commandP,commonParams=requestConfig)

		res = self.uqlSingle(uqlMaker=uqlMaker)
		res.items = res.alias("_nodeSchema").asSchemas()[0]
		return res
		
	
	def getEdgeSchema(self, schemaName: str,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseSchema:
		'''
		Acquire a designated Edge Schema.

		Args:
			schemaName: The name of Schema
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseSchema
		'''

		
		command=CommandList.showEdgeSchema

		if schemaName:
			commandP="@`%s`" % (schemaName)


		uqlMaker = UQLMAKER(command, commandP=commandP,commonParams=requestConfig)

		res = self.uqlSingle(uqlMaker=uqlMaker)
		res.items = res.alias("_edgeSchema").asSchemas()[0]
		return res
	

	def createSchemaIfNotExist(self, schema: Schema,
					 requestConfig: RequestConfig = RequestConfig())->Tuple [bool,ULTIPA_RESPONSE.UltipaResponse]:
		'''
		Create a schema if schema does not exist.

		Args:
			schema: An object of Schema class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		check=self.getSchema(schemaName=schema.name,dbType=schema.DBType)
		if check.status.code==ULTIPA.Code.SUCCESS:
			return [True,ULTIPA_RESPONSE.UltipaResponse()]

		elif check.status.code==ULTIPA.Code.FAILED:
			res=self.createSchema(schema=schema,isCreateProperties=True,requestConfig=requestConfig)
			return [False,res]


	def getProperty(self, dbType: DBType, propertyName: str, schemaName: str , 
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseProperty:
		'''
		Get a property.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			propertyName: The name of the Property

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		res= self.getSchema(schemaName=schemaName,dbType=dbType,requestConfig=requestConfig)
		properties=res.items.properties
		for property in properties:
			if property.name==propertyName:
				return property
		return None	
	
	def getNodeProperty(self,propertyName: str, schemaName: str , 
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseProperty:
		'''
		Get a Node property.

		Args:
			schemaName: The name of schema

			propertyName: The name of the Property

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		res=self.getProperty(propertyName=propertyName,schemaName=schemaName,dbType=DBType.DBNODE,requestConfig=requestConfig)
		return res
	
	def getEdgeProperty(self,propertyName: str, schemaName: str , 
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseProperty:
		'''
		Get an Edge property.

		Args:
			schemaName: The name of schema

			propertyName: The name of the Property

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		res=self.getProperty(propertyName=propertyName,schemaName=schemaName,dbType=DBType.DBEDGE,requestConfig=requestConfig)
		return res
	
	def createPropertyIfNotExist(self, dbType:DBType, schemaName:str, prop:Property,
					 requestConfig: RequestConfig = RequestConfig())->Tuple [bool,ULTIPA_RESPONSE.ResponseProperty]:
		'''
		Create a schema if schema does not exist.

		Args:
			schema: An object of Schema class

			requestConfig: An object of RequestConfig class

		Returns:
			Bool
			ResponseGetProperty

		'''

		res=self.getProperty(dbType=dbType,propertyName=prop.name,schemaName=schemaName,requestConfig=requestConfig)
		if res!= None:
			return[True,ULTIPA_RESPONSE.ResponseProperty()]
		if res == None:
			newprop=PropertyExtra.createProperty(self,dbType=dbType,schemaName=schemaName,prop=prop)
			return [False, newprop]


from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig



# class TemplateExtra(BaseExtra):
# 	def template(self, request: ULTIPA_REQUEST.Template,
# 				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
# 		'''
# 		:EXP: conn.searchNode(ULTIPA_REQUEST.SearchMate(id='1',select=['name']))
# 		:param request: ULTIPA_REQUEST
# 		:return:
# 		'''
# 		command = request.isKhopTemplate and CommandList.khop or CommandList.template
# 		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig, commandP=request.alias)
# 		uqlMaker.addTemplateParams(request.items)
# 		uqlMaker.addParam("limit", request.limit)
# 		uqlMaker.addParam("order_by", request.order_by)
# 		uqlMaker.addParam("return", request._return)
# 		if request.select:
# 			uqlMaker.addParam("select", request.select)
# 		res = self.uqlSingle(uqlMaker)
# 		if res.status.code != ULTIPA.Code.SUCCESS:
# 			return res
# 		return res



from ultipa.structs import DBType
from ultipa.structs.Schema import Schema
from ultipa.types import ULTIPA
from ultipa.utils import common as COMMON
from ultipa.utils.ufilter import ufilter as FILTER
from ultipa.utils.ufilter.new_ufilter import *
from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge
from ultipa.configuration.InsertConfig import InsertConfig




class CommonSchema:
	'''
	    Data calss for schema with name and description.
	'''
	def __init__(self, schema: str, property: str):
		self.schemaName = '@' + schema
		self.propertyName = property

	@property
	def toString(self):
		return "%s.%s" % (self.schemaName, self.propertyName)


class UltipaPath:
	def __init__(self, nodeSchema: List[CommonSchema], edgeSchema: List[CommonSchema]):
		self.nodeSchema = nodeSchema
		self.edgeSchema = edgeSchema

	@property
	def toString(self):
		if self.nodeSchema == '*':
			nodeSchema = '*'
		else:
			nodeSchema = ','.join([i.toString for i in self.nodeSchema])

		if self.edgeSchema == '*':
			edgeSchema = '*'
		else:
			edgeSchema = ','.join([i.toString for i in self.edgeSchema])

		return "{%s}{%s}" % (nodeSchema, edgeSchema)


class Return:
	def __init__(self, alias: str, propertys: List[str] = None, allProperties: bool = False, limit: int = COMMON.LIMIT):
		if propertys is None:
			propertys = []
		self.aliasName = alias
		self.propertys = propertys
		self.all = allProperties
		self.limit = limit

	@property
	def toString(self):
		if self.all:
			return "%s{%s} limit %s" % (self.aliasName, "*", self.limit)
		if len(self.propertys) == 1:
			return "%s.%s limit %s" % (self.aliasName, self.propertys[0], self.limit)
		else:
			return "%s{%s} limit %s" % (self.aliasName, ','.join(self.propertys), self.limit)




class CreateUser:
	def __init__(self, username: str, password: str, graphPrivileges: [dict] = None,
				 systemPrivileges: List[str] = None, propertyPrivileges:dict=None,policies: List[str] = None):
		self.username = username
		self.password = password
		self.graph_privileges = graphPrivileges
		self.system_privileges = systemPrivileges
		self.policies = policies
		self.property_privileges=propertyPrivileges

class AlterUser(CreateUser):
	def __init__(self, username: str, password: str = None, graph_privileges: [dict] = None,
				 system_privileges: List[str] = None, property_privileges:dict=None,policies: List[str] = None):
		super().__init__(username, password, graph_privileges, property_privileges ,system_privileges, policies)


class getUserSetting():
	def __init__(self,username:str,type:str=''):
		self.username=username
		self.type=type

class setUserSetting(getUserSetting):
    def __init__(self, username: str,type:str='',data: str=''):
        super().__init__(username,type)
        self.data = data

# class ShowTask:
# 	def __init__(self, id: int = None, name: str = None, limit: int = None, status: str = ''):
# 		self.id = id
# 		self.limit = limit
# 		self.name = name
# 		self.status = status


# class ClearTask:
# 	def __init__(self, id: int = None, name: str = None, status: str = None, all: bool = False):
# 		self.id = id
# 		self.name = name
# 		self.status = status
# 		self.all = all




# class InsertNodes:
# 	def __init__(self, nodes: List[Node], schemaName: str, config:InsertConfig):
# 		combined_values=[] #to combine values and id for insertion
# 		for node in nodes:
# 			combined_values.append({
# 				"_id":node.id,
# 				**node.values
# 			})
		
# 		self.nodes = combined_values
# 		self.schemaName = '@' + schemaName
# 		self.insertType=config.insertType
# 		self.silent = config.silent

# 	def setSchema(self, schema: str):
# 		self.schemaName = '@' + schema


# class InsertEdges:
# 	def __init__(self, edges: List[Edge], schemaName: str, config:InsertConfig):
# 		combined_values=[] #to combine values and id for insertion
# 		for edge in edges:
# 			combined_values.append({
# 				**edge.values,
# 				'_from':edge.from_id,
# 				'_to':edge.to_id
# 			})
# 		self.edges=combined_values	
# 		self.schemaName = '@' + schemaName
# 		self.insertType=config.insertType
# 		self.silent=config.silent

# 	def setSchema(self, schema: str):
# 		self.schemaName = '@' + schema


class SearchNode:
	def __init__(self, select: Return, id=None,
				 filter: UltipaFilter or list or str = None):
		if id is None:
			id = []
		self.id = id
		self.filter = filter
		self.select = select


class SearchEdge(SearchNode):
	pass


class UpdateNode:
	def __init__(self, values: dict, uuid: [int] = None, filter: UltipaFilter = None, silent: bool = False):
		if uuid is None:
			uuid = []
		self.id = uuid
		self.filter = filter
		self.values = values
		self.silent = silent


class UpdateEdge(UpdateNode):
	pass


class DeleteNode:
	def __init__(self, uuid: [int] = None, filter: UltipaFilter = None, silent: bool = False):
		if uuid is None:
			uuid = []
		self.id = uuid
		self.filter = filter
		self.silent = silent


class DeleteEdge(DeleteNode):
	pass



# class LTE:
# 	def __init__(self, schemaName: CommonSchema, type: DBType):
# 		'''LTE UFE Node and Edge property'''
# 		self.schemaName = schemaName
# 		self.type = type


# class UFE(LTE):
# 	...




class Index(CommonSchema):
	def __init__(self, type: DBType, schema: str, property: str):
		super().__init__(schema=schema, property=property)
		self.DBtype = type



class CreateIndex(Index):
	def __init__(self, type: DBType, schema: str, property: str):
		super().__init__(type, schema, property)


class CreateFulltext(Index):
	def __init__(self, type: DBType, schema: str, property: str, name: str):
		super().__init__(type, schema, property)
		self.name = name


class DropIndex(Index):
	def __init__(self, type: DBType, schema: str, property: str):
		super().__init__(type, schema, property)


class DropFulltext:
	def __init__(self, type: DBType, name: str = ""):
		self.fulltextName = name
		self.DBtype = type




# class Download:
# 	def __init__(self, fileName: str, taskId: str, savePath: str = None):
# 		self.fileName = fileName
# 		self.taskId = taskId
# 		self.savePath = savePath


# class Policy:

# 	def __init__(self, name: str, graphPrivileges: dict = None, systemPrivileges: List[str] = None,
# 				 policies: List[str] = None):
# 		self.name = name
# 		self.graph_privileges = graphPrivileges
# 		self.system_privileges = systemPrivileges
# 		self.policies = policies


# class CreatePolicy(Policy):
# 	pass


# class AlterPolicy(Policy):
# 	pass


class GetPolicy:
	def __init__(self, name: str):
		self.name = name


class DropPolicy(GetPolicy):
	pass


# class GrantPolicy:
# 	def __init__(self, username: str = '', graphPrivileges: dict = None,
# 				 systemPrivileges: List[str] = None, policies: List[str] = None):
# 		self.username = username
# 		self.graph_privileges = graphPrivileges
# 		self.system_privileges = systemPrivileges
# 		self.policies = policies


# class RevokePolicy(GrantPolicy):
# 	pass




class Export:
	def __init__(self, dbType: DBType, limit: int, schemaName: str, selectPropertiesName: List[str]):
		self.dbType = dbType
		self.limit = limit
		self.selectPropertiesName = selectPropertiesName
		self.schemaName = schemaName




class Truncate:
	def __init__(self, graphName: str, dbType: DBType = None, schema: str = None):
		self.dbType = dbType
		self.graphSetName = graphName
		# self.all = allData
		self.schema = schema


class InstallAlgo:
	def __init__(self, configPath: str, soPath: str):
		self.configPath = configPath
		self.soPath = soPath


class InstallExtaAlgo(InstallAlgo):
	...



class Batch:
	Nodes: List[ULTIPA.EntityRow]
	Edges: List[ULTIPA.EntityRow]
	Schema: Schema

	def __init__(self, Schema: Schema = None, Nodes: List[ULTIPA.EntityRow] = None,
				 Edges: List[ULTIPA.EntityRow] = None):
		if Nodes is None:
			Nodes = []
		if Edges is None:
			Edges = []
		self.Nodes = Nodes
		self.Edges = Edges
		self.Schema = Schema

class TaskStatus(Enum):
    All = '*'
    Pending = 'pending'
    Computing = 'computing'
    Writing = 'writing'
    Failed = 'failed'
    Done = 'done'
    Stopped = 'stopped'



# -*- coding: utf-8 -*-
# @Time    : 2023/8/10 14:54
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Retry.py
class Retry:
	'''
	    Processing class that defines settings for retry.
	'''
	def __init__(self, current: int = 0, max: int = 3):
		self.current = current
		self.max = max

class RetryResponse:
	'''
	    Processing class that defines settings for retry response.
	'''
	def __init__(self, canRetry, currentRetry, nextRetry):
		self.canRetry = canRetry
		self.currentRetry = currentRetry
		self.nextRetry = nextRetry


from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
import json
from typing import Union,Optional
from ultipa.utils.convert import convertToAnyObject
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.types.types_request import TaskStatus
from ultipa.utils.convert import convertToTop


class ALGO_RETURN_TYPE:
	ALGO_RETURN_REALTIME = 1
	ALGO_RETURN_WRITE_BACK = 2
	ALGO_RETURN_VISUALIZATION = 4


class TaskExtra(BaseExtra):
	'''
	Processing class that defines settings for task and process related operations.
	'''

	def top(self,
			requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListTop:
		'''
		Top real-time processes.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListTop

		'''
		uqlMaker = UQLMAKER(command=CommandList.top, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker)
		if len(res.data)>0:
			res.data=convertToTop(res)
		return res

	def kill(self, processId: str = None, all: bool = False,
			 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Kill real-time processes.

		Args:
			id: The ID of real-time process

			all: Whether to kill all real-time processes

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		commonP = []
		if processId:
			commonP = processId
		if all:
			commonP = '*'
		uqlMaker = UQLMAKER(command=CommandList.kill, commonParams=requestConfig)
		uqlMaker.setCommandParams(commonP)
		res = self.uqlSingle(uqlMaker)
		return res

	def showTask(self, algoNameOrId: Optional[Union[int, str]] = None, status: TaskStatus=None,
				 config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListTask:
		'''
		Show back-end tasks.

		Args:

        	algoNameOrld (str): The name of the task algo to show

        	status (str): The status of the tasks to retrieve.

        	config (RequestConfig): An object of RequestConfig class.

		Returns:
			ResponseListTask

		'''

		_jsonKeys = ['taskJson']
		uqlMaker = UQLMAKER(command=CommandList.showTask, commonParams=config)
		commonP = []

		# if all:
		# 	commonP.append('*')

		# else:	

		if isinstance(algoNameOrId, str) or algoNameOrId is None:
			if algoNameOrId and status:
					commonP.append(algoNameOrId)
					commonP.append(status.value)
			if algoNameOrId and not status:
					commonP.append(algoNameOrId)
					commonP.append('*')
			if not algoNameOrId and status:
					commonP.append('*')
					commonP.append(status.value)
		
		elif isinstance(algoNameOrId, int):
				commonP=algoNameOrId

		uqlMaker.setCommandParams(commandP=commonP)
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(jsonKeys=_jsonKeys))
		newDatas = []
		if res.data:
			for obj in res.data:
				obj = obj.__dict__
				newData = ULTIPA_RESPONSE.Task()
				taskJson = obj.get("taskJson", {})
				newData.param = json.loads(taskJson.get("param", "{}"))
				newData.result = taskJson.get("result")
				task_info = taskJson.get("task_info", {})

				if task_info.get('status_code'):
					task_info["status_code"] = ULTIPA.TaskStatusString[task_info.get("TASK_STATUS")]
				if task_info.get('engine_cost'):
					task_info["engine_cost"] = task_info.get("writing_start_time", 0) - task_info.get("start_time", 0)

				newData.task_info = convertToAnyObject(task_info)
				return_type_get = int(task_info.get('return_type', 0))
				return_type = ULTIPA_RESPONSE.Return_Type()
				return_type.is_realtime = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_REALTIME else False
				return_type.is_visualization = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_VISUALIZATION else False
				return_type.is_wirte_back = True if return_type_get & ALGO_RETURN_TYPE.ALGO_RETURN_WRITE_BACK else False
				newData.task_info.__setattr__('return_type', return_type)
				newDatas.append(newData)
			res.data = newDatas
		return res

	def clearTask(self, algoNameOrId: Optional[Union[int, str]] = None, status:TaskStatus=None,
				  config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Clear back-end tasks.

		Args:

        	algoNameOrId (str): The name of the algotask to clear.

        	status (str): The status of the tasks to clear.

        	requestConfig (RequestConfig): An object of RequestConfig class.

		Returns:
			Response

		'''

		uqlMaker = UQLMAKER(command=CommandList.clearTask, commonParams=config)
		commonP = []
		# if all:
		# 	commonP.append('*')
		# else:


		if isinstance(algoNameOrId, str) or algoNameOrId is None:
			if algoNameOrId and status:
					commonP.append(algoNameOrId)
					commonP.append(status.value)
			if algoNameOrId and not status:
					commonP.append(algoNameOrId)
					commonP.append('*')
			if not algoNameOrId and status:
					commonP.append('*')
					commonP.append(status.value)
		
		elif isinstance(algoNameOrId, int):
				commonP = algoNameOrId
		# if id:
		# 		commonP = id

		uqlMaker.setCommandParams(commandP=commonP)
		res = self.UqlUpdateSimple(uqlMaker)

		return res

	def stopTask(self, id: str = None,
				 config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Stop back-end tasks.

		Args:
			id: The ID of back-end task

			all: Whether to stop all back-end tasks that are computing

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		uqlMaker = UQLMAKER(command=CommandList.stopTask, commonParams=config)
		commonP = []
		# if all:
		# 	commonP = '*'
		# if id:
		commonP = id
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)
	

	def pauseTask(self, id: int = None, all: bool = False,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		'''
		Stop back-end tasks.

		Args:
			id: The ID of back-end task

			all: Whether to pause all back-end tasks that are computing

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseCommon

		'''
		uqlMaker = UQLMAKER(command=CommandList.pauseTask, commonParams=requestConfig)
		commonP = []
		if all:
			commonP = '*'
		if id:
			commonP = id
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)
	
	def resumeTask(self, id: int = None, all: bool = False,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseCommon:
		'''
		Stop back-end tasks.

		Args:
			id: The ID of back-end task

			all: Whether to resume all back-end tasks

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseCommon

		'''
		uqlMaker = UQLMAKER(command=CommandList.resumeTask, commonParams=requestConfig)
		commonP = []
		if all:
			commonP = '*'
		if id:
			commonP = id
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)
	


	def clusterInfo(self,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ClusterInfo:
		'''
		Show cluster information.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ClusterInfo

		'''
		self.refreshRaftLeader(redirectHost='', requestConfig=requestConfig)
		result = []
		if not requestConfig.graphName:
			graphSetName = 'default'
		else:
			graphSetName = requestConfig.graphName
		for peer in self.hostManagerControl.getAllHostStatusInfo(graphSetName):
			info = ULTIPA_RESPONSE.Cluster()
			info.status = peer.status
			info.host = peer.host
			info.isLeader = peer.isLeader
			info.isFollowerReadable = peer.isFollowerReadable
			info.isAlgoExecutable = peer.isAlgoExecutable
			info.isUnset = peer.isUnset
			info.cpuUsage = None
			info.memUsage = None
			if peer.status:
				ret = self.stats(requestConfig=RequestConfig(host=peer.host))
				if ret.status.code == ULTIPA.Code.SUCCESS:
					info.cpuUsage = ret.data.cpuUsage
					info.memUsage = ret.data.memUsage

			result.append(info)
		res = ULTIPA_RESPONSE.Response()
		res.data = result
		return res



# -*- coding: utf-8 -*-
# @Time    : 2024/7/5 18:14
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : InsertRequestConfig.py
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType
from ultipa.structs.Retry import Retry


class InsertRequestConfig(RequestConfig):
	'''
    Configuration class for the interface of insert operation.

    This class provides settings for inserting metadata.
    
    Args:
        - insertType (InsertType): The mode of insert request (NORMAL/UPSERT/OVERWRITE).
        - graphName (str): The name of graphset to use.
        - timeout (int): The timeout for the request in seconds.
        - retry (Retry): The object of Retry configuration classwhen the request fails
        - stream (bool): Whether to return stream
        - useHost (str): The designated host the request will be sent to, or sent to a random host if not set
        - useMaster (bool):	Whether to send the request to the leader to guarantee Consistency Read
        - CreateNodeIfNotExist (bool): Whether to create start/end nodes of edge when these nodes do not exist in the graphset
        - timeZone (str): The string of timezone in standard format
        - timeZoneOffset (any): 1, the number of seconds; 2, a 5-character string such as +0700, -0430
	'''
	def __init__(self, insertType: InsertType, graphName: str = '', timeout: int = 3600,
				 retry: Retry = Retry(), stream: bool = False, useHost: str = None, useMaster: bool = False,
				 CreateNodeIfNotExist: bool = False, timeZone=None, timeZoneOffset=None, **kwargs):
		super().__init__(graphName, timeout, retry, stream, useHost, useMaster, timeZone=timeZone,
						 timeZoneOffset=timeZoneOffset)
		self.insertType = insertType
		if kwargs.get("silent") is not None:
			self.silent = kwargs.get("silent")
		else:
			self.silent = True
		if kwargs.get("batch") is not None:
			self.batch = kwargs.get("batch")
		if kwargs.get("n") is not None:
			self.n = kwargs.get("n")
		if kwargs.get("timeoutWithSeconds") is not None:
			self.timeoutWithSeconds = kwargs.get("timeoutWithSeconds")
		self.createNodeIfNotExist = CreateNodeIfNotExist



from ultipa.operate.base_extra import BaseExtra

from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.InsertType import InsertType
from ultipa.configuration.InsertConfig import InsertConfig
from ultipa.configuration.InsertRequestConfig import InsertRequestConfig
from ultipa.structs.Edge import Edge
from ultipa.utils.ufilter.new_ufilter import *


class EdgeExtra(BaseExtra):
	'''
	Processing class that defines settings for edge related operations.

	'''

	def searchEdge(self, request: ULTIPA_REQUEST.SearchEdge,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseSearchEdge:
		'''
		Query for edges.

		Args:
			request: An object of SearchEdge class 

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseSearchEdge

		'''

		uqlMaker = UQLMAKER(command=CommandList.edges, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)

		uqlMaker.addParam('as', request.select.aliasName)
		uqlMaker.addParam("return", request.select)

		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		return res

	def insertEdges(self,edges: List[Edge], schemaName: str, config:InsertRequestConfig) -> ULTIPA_RESPONSE.ResponseInsertEdge:
		'''
		Insert edges.

		Args:
			edges: List of edges to be inserted

			schemaName: Name of the schema

			config: An object of InsertRequestConfig class

		Returns:
			ResponseInsertEdge
		'''
		combined_values = []  # to combine values and id for insertion
		for edge in edges:
			edge_dict = {}
			if edge.from_id:
				edge_dict["_from"] = edge.from_id
			if edge.to_id:
				edge_dict["_to"]= edge.to_id

			if edge.from_uuid:
				edge_dict["_from_uuid"]	=edge.from_uuid
			if edge.to_uuid:
				edge_dict["_to_uuid"]=edge.to_uuid

			if edge.uuid:
				edge_dict["_uuid"]=edge.uuid		
			edge_dict.update(edge.values)
			combined_values.append(edge_dict)
		edges=combined_values
		schemaName = '@' + schemaName	

		uqlMaker = UQLMAKER(command=CommandList.insert, commonParams=config)
		if config.insertType==InsertType.UPSERT:
			uqlMaker = UQLMAKER(command=CommandList.upsert, commonParams=config)
		if config.insertType==InsertType.OVERWRITE:
			uqlMaker.addParam('overwrite', "", required=False)
		if schemaName:
			uqlMaker.addParam('into', schemaName, required=False)
		uqlMaker.addParam('edges', edges)

		if config.silent==False:
			uqlMaker.addParam('as', "edges")
			# uqlMaker.addParam('return', "edges._uuid")
			uqlMaker.addParam('return', "edges{*}")
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		if config.silent==False:
			if len(res.aliases) > 0:
				res.data = res.items.get(res.aliases[0].alias).data
		return res

	def updateEdges(self, request: ULTIPA_REQUEST.UpdateEdge,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Update edges.

		Args:
			request: An object of UpdateEdge class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.updateEdges, commonParams=requestConfig)
		if request.id:
			uqlMaker.setCommandParams(request.id)
		elif request.filter:
			uqlMaker.setCommandParams(request.filter)
		uqlMaker.addParam("set", request.values)
		uqlMaker.addParam("silent", request.silent)
		res = self.UqlUpdateSimple(uqlMaker)
		return res

	def deleteEdges(self,filter: str, config:InsertRequestConfig) -> ULTIPA_RESPONSE.ResponseDeleteEdge:
		'''
		Delete edges.

		Args:
			filtert: An object of UltipaFilter class
			
			config: An object of InsertConfig class

		Returns:
			ResponseDeleteEdge

		'''

		uqlMaker = UQLMAKER(command=CommandList.deleteEdges, commonParams=config)
		# if request.id:
		# 	uqlMaker.setCommandParams(request.id)
		if filter:
			uqlMaker.setCommandParams(filter)

		if config.silent==False:
			uqlMaker.addParam('as', "edges")
			# uqlMaker.addParam('return', "edges._uuid")
			uqlMaker.addParam('return', "edges{*}")

		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			return res
		if config.silent==False:
			if len(res.aliases) > 0:
				res.data = res.items.get(res.aliases[0].alias).data
		return res		



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 11:59
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : commonUql.py

class GetProperty():
	node: str = 'show().node_property()'
	edge: str = 'show().edge_property()'


class GetPropertyBySchema():
	node: str = 'show().node_schema()'
	edge: str = 'show().edge_schema()'


# -*- coding: utf-8 -*-
# @Time    : 2023/12/15 14:49
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : PointType.py
class PointType:

	@staticmethod
	def PointFunction(latitude: float, longitude: float):
		return "point({latitude:%s, longitude:%s})" % (latitude, longitude)

	@staticmethod
	def PointString(latitude: float, longitude: float):
		return f"POINT({latitude} {longitude})"



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 18:15
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : InsertType.py
from ultipa.proto import ultipa_pb2


class InsertType:
	'''
	    Data class for insert type.
	'''
	NORMAL = ultipa_pb2.NORMAL
	OVERWRITE = ultipa_pb2.OVERWRITE
	UPSERT = ultipa_pb2.UPSERT


# -*- coding: utf-8 -*-
# @Time    : 2024/05/17 10:56
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Stats.py
from ultipa.structs.BaseModel import BaseModel


class Stats(BaseModel):
	'''
	    Data class for Statistics.
	'''

	def __init__(self, cpuUsage: str, memUsage: str,company: str,cpuCores: str,expiredDate: str,serverType: str,version: str):
		self.cpuUsage=cpuUsage
		self.memUsage=memUsage
		self.company=company
		self.cpuCores=cpuCores
		self.expiredDate=expiredDate
		self.serverType=serverType
		self.version=version
	
	
	
	
	
	
	


from setuptools import setup, find_packages

APP = ['ultipa']
DATA_FILES = []
OPTIONS = {}

def readMe():
    try:

        ret = open("""D:\work\\app\\pythonProject\\MakeUltipaPackage\\ReadMe.md""", encoding="utf-8").read()
    except Exception as e:
        return ""
    return ret

setup(
    app=APP,
    name="ultipa",
    metaversion="",
    version="4.5.1",
    python_requires='>=3.6, <=3.10',
    packages=find_packages(),  # 常用,要熟悉 :会自动查找当前目录下的所有模块(.py文件) 和包(包含__init___.py文件的文件夹)
    # scripts = ['say_hello.py'],
    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=[
                'grpcio==1.48.2',
                'grpcio-tools==1.48.2',
                'protobuf==3.19.0',
                'google==2.0.3',
                'schedule==1.1.0',
                'prettytable==2.5.0',
                'treelib==1.6.1',
                "tzlocal==4.2",
                "pytz==2022.7",
                "future~=0.18.2",
                "python-dateutil~=2.8.2"
                ],  # 常用
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.rst',"printer"],
        # And include any *.msg files found in the 'hello' package, too:
        'hello': ['*.msg'],
    },
    # metadata for upload to PyPI
    author="Ultipa",
    author_email="support@ultipa.com",
    description="Pure Python Ultipa Driver",
    license="PSF",
    keywords="ultipa sdk,ultipa graph",
    url="https://www.ultipa.com/document/ultipa-drivers/python-installation",  # project home page, if any
    long_description=readMe(),
    long_description_content_type='text/markdown',
    # could also include long_description, download_url, classifiers, etc.
)




from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.configuration.RequestConfig import RequestConfig
from typing import Callable,List
from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge
class ExportExtra(BaseExtra):
	'''
		Processing class that defines settings for data exporting operation.
	'''

	def export(self, request: ULTIPA_REQUEST.Export,cb: Callable[[List[Node], List[Edge]], None],
			   requestConfig: RequestConfig = RequestConfig()) :
		'''
		Export data.

		Args:
			request: An object of Export class

			requestConfig: An object of RequestConfig class

		Returns:
			Stream
		'''
		self.exportData(request,cb,requestConfig)
		return 



import copy
import json
from typing import List

from ultipa.proto.ultipa_pb2 import AttrListData
from ultipa.structs import Property
from ultipa.types.types import ResultType
from ultipa.utils import errors
from ultipa.types import ULTIPA, ULTIPA_RESPONSE, ULTIPA_REQUEST
from ultipa.utils.errors import ParameterException
from ultipa.utils.serialize import _Serialize
from ultipa.utils.typeCheck import TypeCheck


class HasDataMap:
	has_ultipa_data: bool
	has_attr_data: bool
	only_attr_list: bool

	def __init__(self, has_ultipa_data, has_attr_data, only_attr_list):
		self.has_ultipa_data = has_ultipa_data
		self.has_attr_data = has_attr_data
		self.only_attr_list = only_attr_list


class FormatResponse:

	@staticmethod
	def resTableToArray(table: ULTIPA.Table):
		tres = []
		headers = []
		if table:
			try:
				for index, header in enumerate(table.headers):
					if not header.property_name:
						if header.property_type == ULTIPA.PropertyType.PROPERTY_NULL:
							header.property_name = f"null({index})"
					headers.append({"property_name": header.property_name,
									"property_type": Property._getStringByPropertyType(
										header.property_type)})
				tret = ULTIPA.Table(table_name=table.name, headers=headers, table_rows=table.rows)
				tres.append(tret)
			except Exception as e:
				raise errors.ParameterException(e)
		return tret

	@staticmethod
	def formatStatisticsTable(table: ULTIPA.Table):
		headers = []
		if table:
			try:
				for header in table.headers:
					headers.append({"property_name": header.property_name,
									"property_type": Property._getStringByPropertyType(
										header.property_type)})
				item = {}
				for row in table.rows:
					for index in range(len(row)):
						v = row[index]
						item[table.headers[index].property_name] = int(v)
				tret = ULTIPA.UltipaStatistics(edge_affected=item.get("edge_affected"),
											   node_affected=item.get("node_affected"),
											   total_time_cost=item.get("total_time_cost"),
											   engine_time_cost=item.get("engine_time_cost"))
				return tret
			except Exception as e:
				raise errors.ParameterException(e)


class DataMerge:
	@staticmethod
	def arrayDiff(arr1: "[str]", arr2: "[str]"):
		diff = []
		for i in arr2:
			if i not in arr1:
				diff.append(i)
		return diff

	@staticmethod
	def customizer(objValue, srcValue, key, ignoreKeys):
		if ignoreKeys and key in ignoreKeys:
			return objValue

		if isinstance(objValue, list):
			return objValue + srcValue

	@staticmethod
	def merge(o1: object, o2: object, ignoreKeys: [str] = None):
		if ignoreKeys is None:
			ignoreKeys = []
		retDict = {}
		for key in o1:
			if o1.get(key) and o2.get(key):
				mergeRet = DataMerge.customizer(o1[key], o2[key], key, ignoreKeys)
				retDict.update({key: mergeRet})
			else:
				if isinstance(o1, dict) and isinstance(o1, dict):
					if o1.get(key):
						retDict.update({key: o1.get(key)})
						retDict.update(o2)
		return retDict

	@staticmethod
	def concat(arr1: list, arr2: list, bykey: str = None, ignoreKeys: "[str]" = None):

		if bykey:
			if bykey == 'table_name':
				if isinstance(arr1, list) and isinstance(arr2, list):
					newData = arr1
					newArry = arr2

					# Find the different table_name of arr1 and arr2
					if isinstance(newData, dict):
						diff = DataMerge.arrayDiff([i[bykey] for i in arr1], [i[bykey] for i in arr2])
					else:
						diff = DataMerge.arrayDiff([y.name for y in newData], [y.name for y in arr2])

					# If existing, add the table_name to newData
					if diff:
						for di in diff:
							for a2 in arr2:
								if a2.name == di:
									newData.append(ULTIPA.Table(table_name=di, table_rows=a2.rows, headers=a2.headers))

					# Traverse newArry, add table to table_rows in accordance with table_name  
					for a2 in newArry:
						for data in newData:
							if a2.name == data.name:
								data.rows += a2.rows
					return newData

			if bykey == 'nodes':
				if isinstance(arr1, list) and isinstance(arr2, list):
					newArry = arr2
					newData = arr1

					arrtList = []
					for new in newData:
						arrtList.append(new.alias)
					for arr in newArry:
						if arr.alias not in arrtList:
							newData.append(arr)
					for arr in newArry:
						if not newData:
							newData.append(arr)
						for new in newData:
							if arr.alias == new.alias:
								new.nodes += arr.nodes
					return newData

			if bykey == 'edges':
				if isinstance(arr1, list) and isinstance(arr2, list):
					newArry = arr2
					newData = arr1

					arrtlist = []
					for new in newData:
						arrtlist.append(new.alias)

					for arr in newArry:
						if arr.alias not in arrtlist:
							newData.append(arr)
							return newData

					for arr in newArry:
						if not newData:
							newData.append(arr)
						for new in newData:
							if arr.alias == new.alias:
								new.edges += arr.edges

					return newData

			if bykey == 'values':
				if isinstance(arr1, list) and isinstance(arr2, list):
					newArry = arr2
					newData = arr1
					arrtlist = []
					for new in newData:
						arrtlist.append(new.alias)

					for arr in newArry:
						if arr.alias not in arrtlist:
							newData.append(arr)
					for arr in newArry:
						if not newData:
							newData.append(arr)
						for new in newData:
							if arr.alias in arrtlist and arr.alias == new.alias:
								new.values += arr.values
					return newData

			if bykey == 'arrays':
				if isinstance(arr1, list) and isinstance(arr2, list):
					newArry = arr2
					newData = arr1
					arrtlist = []
					for new in newData:
						arrtlist.append(new.alias)

					for arr in newArry:
						if arr.alias not in arrtlist:
							newData.append(arr)

					for arr in newArry:
						if not newData:
							newData.append(arr)
						for new in newData:
							if arr.alias in arrtlist and arr.alias == new.alias:
								new.elements += arr.elements
					return newData

		if isinstance(arr1, list) and isinstance(arr2, list):
			arr1.extend(arr2)
			return arr1
		elif isinstance(arr1, dict) and isinstance(arr2, dict):
			return arr1.update(arr2)


class FormatType:

	@staticmethod
	def checkProperty(row, schemaProperties):
		propertyList = []
		rowPropertyList = row.values.keys()
		for property in rowPropertyList:
			findSchema = list(filter(lambda x: x.get('name') == property, schemaProperties))
			if findSchema:
				schemaProperty = findSchema[0]
				reqProperty = Property(schemaProperty.get("name"))
				reqProperty.setTypeInt(schemaProperty.get("type"))
				propertyList.append(reqProperty)
			else:
				raise ParameterException(
					err=f"row [{row._getIndex()}] error: schema[{row.schema}] doesn't contain property [{property}].")
		return propertyList

	@staticmethod
	def checkNodeRow(i: int, schema: ULTIPA_REQUEST.Schema, row: ULTIPA.EntityRow):
		if row._index != None:
			i = row._index

		if row is None:
			raise ParameterException(err=f"The row [{i}] data is null")

		if row.values is None:
			raise ParameterException(err=f"node row [{i}] error: values are empty but  properties size > 0.")

		if not schema.properties and row.values:
			raise ParameterException(err=f"node row [{i}] error: properties are empty but values size > 0.")

		if len(row.values.keys()) > len(schema.properties):
			raise ParameterException(err=f"node row [{i}] error: values size larger than properties size.")

		if len(row.values.keys()) < len(schema.properties):
			raise ParameterException(err=f"node row [{i}] error: values size smaller than properties size.")

		for proper in schema.properties:
			if proper.isIdType() or proper.isIgnore():
				continue
			if proper.name not in row.values:
				raise ParameterException(
					err=f"node row [{i}] error: values doesn't contain property [{proper.name}].")

	@staticmethod
	def checkEdgeRow(i: int, schema: ULTIPA_REQUEST.Schema, row: ULTIPA.EntityRow):
		if row._index != None:
			i = row._index

		if row is None:
			raise ParameterException(err=f"The row [{i}] data is null")
		if row.values is None:
			raise ParameterException(err=f"edge row [{i}] error: values are empty but properties size > 0.")
		if not schema.properties and row.values:
			raise ParameterException(err=f"edge row [{i}] error: properties are empty but values size > 0.")

		if len(row.values.keys()) > len(schema.properties):
			raise ParameterException(err=f"edge row [{i}] error: values size larger than properties size.")

		if len(row.values.keys()) < len(schema.properties):
			raise ParameterException(err=f"edge row [{i}] error: values size smaller than properties size.")

		if ((row.from_id is None or row.from_id == "") and (row.to_id is None or row.to_id == "")) \
				and (row.from_uuid is None and row.to_uuid is None):
			raise ParameterException(err=f"row [{i}] error: _from/_from_uuid and _to/_to_uuid are null.")

		if (row.from_id != None and row.from_id != "") and (row.to_id is None or row.to_id == ""):
			raise ParameterException(err=f"row [{i}] error: _from has value [{row.from_id}] but _to got null.")

		if (row.from_id is None or row.from_id == "") and (row.to_id != None and row.to_id != ""):
			raise ParameterException(err=f"row [{i}] error: _to has value [{row.to_id}] but _from got null.")

		if row.from_uuid is None and row.to_uuid != None:
			raise ParameterException(
				err=f"row [{i}] error: _to_uuid has value [{row.to_uuid}] but _from_uuid got null.")

		if row.from_uuid != None and row.to_uuid is None:
			raise ParameterException(
				err=f"row [{i}] error: _from_uuid has value [{row.from_uuid}] but _to_uuid got null.")

		for proper in schema.properties:
			if proper.isIdType() or proper.isIgnore():
				continue
			if proper.name not in row.values:
				raise ParameterException(
					err=f"edge row [{i}] error: values doesn't contain property [{proper.name}].")

	@staticmethod
	def checkEntityNodeRow(i: int, schema: ULTIPA_REQUEST.Schema, row: ULTIPA.EntityRow):
		if row._index != None:
			i = row._index

		if row is None:
			raise ParameterException(err=f"The row [{i}] data is null")

		if row.values is None:
			raise ParameterException(err=f"node row [{i}] error: values are empty but  properties size > 0.")

		if not schema.properties and row.values:
			raise ParameterException(err=f"node row [{i}] error: properties are empty but values size > 0.")

		if len(row.values.keys()) > len(schema.properties):
			raise ParameterException(err=f"node row [{i}] error: values size larger than properties size.")

		if len(row.values.keys()) < len(schema.properties):
			raise ParameterException(err=f"node row [{i}] error: values size smaller than properties size.")

		for proper in schema.properties:
			if proper.isIdType() or proper.isIgnore():
				continue
			if proper.name not in row.values:
				raise ParameterException(
					err=f"node row [{i}] error: values doesn't contain property [{proper.name}].")

	@staticmethod
	def checkEntityEdgeRow(i: int, schema: ULTIPA_REQUEST.Schema, row: ULTIPA.EntityRow):
		if row._index != None:
			i = row._index

		if row is None:
			raise ParameterException(err=f"The row [{i}] data is null")
		if row.values is None:
			raise ParameterException(err=f"edge row [{i}] error: values are empty but properties size > 0.")
		if not schema.properties and row.values:
			raise ParameterException(err=f"edge row [{i}] error: properties are empty but values size > 0.")

		if len(row.values.keys()) > len(schema.properties):
			raise ParameterException(err=f"edge row [{i}] error: values size larger than properties size.")

		if len(row.values.keys()) < len(schema.properties):
			raise ParameterException(err=f"edge row [{i}] error: values size smaller than properties size.")

		if ((row.from_id is None or row.from_id == "") and (row.to_id is None or row.to_id == "")) \
				and (row.from_uuid is None and row.to_uuid is None):
			raise ParameterException(err=f"row [{i}] error: _from/_from_uuid and _to/_to_uuid are null.")

		if (row.from_id != None and row.from_id != "") and (row.to_id is None or row.to_id == ""):
			raise ParameterException(err=f"row [{i}] error: _from has value [{row.from_id}] but _to got null.")

		if (row.from_id is None or row.from_id == "") and (row.to_id != None and row.to_id != ""):
			raise ParameterException(err=f"row [{i}] error: _to has value [{row.to_id}] but _from got null.")

		if row.from_uuid is None and row.to_uuid != None:
			raise ParameterException(
				err=f"row [{i}] error: _to_uuid has value [{row.to_uuid}] but _from_uuid got null.")

		if row.from_uuid != None and row.to_uuid is None:
			raise ParameterException(
				err=f"row [{i}] error: _from_uuid has value [{row.from_uuid}] but _to_uuid got null.")

		for proper in schema.properties:
			if proper.isIdType() or proper.isIgnore():
				continue
			if proper.name not in row.values:
				raise ParameterException(
					err=f"edge row [{i}] error: values doesn't contain property [{proper.name}].")

	@staticmethod
	def status(_status, host: str = None) -> ULTIPA.Status:
		statusData = ULTIPA.Status(code=_status.error_code, message=_status.msg)
		isNotRaftMode = _status.error_code == ULTIPA.Code.NOT_RAFT_MODE

		if _status.cluster_info or isNotRaftMode:
			leaderPeer = None
			_followrs = []
			if _status.cluster_info.leader_address or host:
				leaderPeer = ULTIPA.RaftPeerInfo(_status.cluster_info.leader_address or host, True, True, False, False,
												 False)
				leaderPeer.isAlgoExecutable = isNotRaftMode or False
				_followrs.append(leaderPeer)
			for foll in _status.cluster_info.followers:
				status = foll.status == 1
				peer = ULTIPA.RaftPeerInfo(foll.address)
				peer.status = status
				peer.isLeader = False
				peer.isAlgoExecutable = False
				peer.isFollowerReadable = False
				if status:
					peer.isAlgoExecutable = foll.role & ULTIPA.FollowerRole.ROLE_ALGO_EXECUTABLE and True or False
					peer.isFollowerReadable = foll.role & ULTIPA.FollowerRole.ROLE_READABLE and True or False
				peer.isUnset = foll.role == ULTIPA.FollowerRole.ROLE_UNSET
				_followrs.append(peer)
			clusterInfo = ULTIPA.ClusterInfo(_status.cluster_info.redirect, _followrs, leader=leaderPeer)
			statusData.clusterInfo = clusterInfo
		return statusData

	@staticmethod
	def getRaftStatus(data: dict) -> ULTIPA.Status:
		statusData = ULTIPA.Status(code=data.get("code"), message=data.get("message"))
		clusterInfo = ULTIPA.ClusterInfo(redirect=data.get("redirectHost"), raftPeers=data.get("followersPeerInfos"),
										 leader=data.get("leaderInfos"))
		statusData.clusterInfo = clusterInfo
		return statusData

	@staticmethod
	def table(_table, timeZone, timeZoneOffset) -> ULTIPA.Table:
		table_name = _table.table_name
		headers = []
		for h in _table.headers:
			headers.append(h)
		values = []
		for index, row in enumerate(_table.table_rows):
			_vs = []
			for index2, _v in enumerate(row.values):
				if len(headers) > 0:
					property_type = headers[index2].property_type
				else:
					property_type = ULTIPA.PropertyType.PROPERTY_STRING
				_seria = _Serialize(type=property_type, value=_v, timeZone=timeZone, timeZoneOffset=timeZoneOffset)
				unret = _seria.unserialize()
				_vs.append(unret)
			values.append(_vs)
		tableData = ULTIPA.Table(table_name=table_name, headers=headers, table_rows=values)
		return tableData

	@staticmethod
	def tables(_tables, timeZone, timeZoneOffset) -> List[ULTIPA.Table]:
		tablesData = []
		if _tables:
			for table in _tables:
				tableData = FormatType.table(table, timeZone, timeZoneOffset)
				tret = FormatResponse.resTableToArray(table=tableData)
				tablesData.append(tret)
		return tablesData

	@staticmethod
	def statistics(_table, timeZone, timeZoneOffset) -> ULTIPA.UltipaStatistics:
		if not _table:
			return None
		tableData = FormatType.table(_table, timeZone, timeZoneOffset)
		tret = FormatResponse.formatStatisticsTable(table=tableData)
		return tret

	@staticmethod
	def nodeTable(nodeTableData, timeZone, timeZoneOffset):
		SchemaTypeDict = {}
		SchemaHeaderDict = {}
		SchemaSubTypeHeaderDict = {}
		tables = []
		for header in nodeTableData.node_table.schemas:
			schemaName = header.schema_name
			typesDict = {}
			headerDict = {}
			subTypesDict = {}
			for index, header in enumerate(header.properties):
				typesDict.update({index: header.property_type})
				headerDict.update({index: header.property_name})
				subTypesDict.update({index: header.sub_types})

			SchemaTypeDict.update({schemaName: typesDict})
			SchemaHeaderDict.update({schemaName: headerDict})
			SchemaSubTypeHeaderDict.update({schemaName: subTypesDict})

		schemaTypeGet = SchemaTypeDict.get
		schemaHeaderGet = SchemaHeaderDict.get
		schemaSubTypeGet = SchemaSubTypeHeaderDict.get

		for node_row in nodeTableData.node_table.entity_rows:
			if node_row.is_null is True:
				tables.append(None)
				continue
			data = {}
			_id = node_row.id
			uuid = node_row.uuid

			schema_name = node_row.schema_name
			if node_row.values:
				for index, uvalue in enumerate(node_row.values):
					seria = _Serialize(type=schemaTypeGet(schema_name).get(index), value=uvalue,
									   subTypes=schemaSubTypeGet(schema_name).get(index), timeZone=timeZone,
									   timeZoneOffset=timeZoneOffset)
					value = seria.unserialize()
					data.update({schemaHeaderGet(schema_name).get(index): value})
			node = ULTIPA.Node(id=_id, values=data, schema_name=schema_name, uuid=uuid)
			tables.append(node)
		return tables

	@staticmethod
	def edgeTable(edgeTableData, timeZone, timeZoneOffset):
		SchemaTypeDict = {}
		SchemaHeaderDict = {}
		SchemaSubTypeHeaderDict = {}
		tables = []
		for header in edgeTableData.edge_table.schemas:
			schemaName = header.schema_name
			typesDict = {}
			headerDict = {}
			subTypesDict = {}
			for index, header in enumerate(header.properties):
				typesDict.update({index: header.property_type})
				headerDict.update({index: header.property_name})
				subTypesDict.update({index: header.sub_types})

			SchemaTypeDict.update({schemaName: typesDict})
			SchemaHeaderDict.update({schemaName: headerDict})
			SchemaSubTypeHeaderDict.update({schemaName: subTypesDict})

		schemaTypeGet = SchemaTypeDict.get
		schemaHeaderGet = SchemaHeaderDict.get
		schemaSubTypeGet = SchemaSubTypeHeaderDict.get

		for edge_row in edgeTableData.edge_table.entity_rows:
			if edge_row.is_null is True:
				tables.append(None)
				continue
			data = {}
			_from_uuid = edge_row.from_uuid
			_from_id = edge_row.from_id
			_to_uuid = edge_row.to_uuid
			_to_id = edge_row.to_id
			schema_name = edge_row.schema_name
			uuid = edge_row.uuid
			if edge_row.values:
				for index, uvalue in enumerate(edge_row.values):
					seria = _Serialize(type=schemaTypeGet(schema_name).get(index), value=uvalue,
									   subTypes=schemaSubTypeGet(schema_name).get(index), timeZone=timeZone,
									   timeZoneOffset=timeZoneOffset)
					value = seria.unserialize()
					data.update({schemaHeaderGet(schema_name).get(index): value})
			edge = ULTIPA.Edge(from_id=_from_id, to_id=_to_id, values=data, uuid=uuid, from_uuid=_from_uuid,
							   to_uuid=_to_uuid, schema_name=schema_name)
			tables.append(edge)
		return tables

	@staticmethod
	def attrEntity(entityRows, type: str):
		entityList = []
		for data in entityRows.entity_rows:
			if type == "node":
				entityList.append(ULTIPA.Node(values=None, uuid=data.uuid))
			else:
				entityList.append(ULTIPA.Edge(values=None, uuid=data.uuid))
		return entityList

	@staticmethod
	def propertyType(type: str) -> ULTIPA.PropertyType:
		'''
		Convert the type input by the user to ULTIPA.PropertyType
		:param type:
		:return:
		'''
		if type.upper() == 'PROPERTY_STRING' or type.upper() == 'STRING':
			type = ULTIPA.PropertyType.PROPERTY_STRING
		elif type.upper() == 'PROPERTY_INT32' or type.upper() == 'INT32':
			type = ULTIPA.PropertyType.PROPERTY_INT32
		elif type.upper() == 'PROPERTY_INT64' or type.upper() == 'INT64':
			type = ULTIPA.PropertyType.PROPERTY_INT64
		elif type.upper() == 'PROPERTY_FLOAT' or type.upper() == 'FLOAT':
			type = ULTIPA.PropertyType.PROPERTY_FLOAT
		elif type.upper() == 'PROPERTY_DOUBLE' or type.upper() == 'DOUBLE':
			type = ULTIPA.PropertyType.PROPERTY_DOUBLE
		elif type.upper() == 'PROPERTY_UINT32' or type.upper() == 'UINT32':
			type = ULTIPA.PropertyType.PROPERTY_UINT32
		elif type.upper() == 'PROPERTY_UINT64' or type.upper() == 'UINT64':
			type = ULTIPA.PropertyType.PROPERTY_UINT64
		elif type.upper() == 'PROPERTY_DATETIME' or type.upper() == 'DATETIME':
			type = ULTIPA.PropertyType.PROPERTY_DATETIME
		elif type.upper() == 'PROPERTY_TIMESTAMP' or type.upper() == 'TIMESTAMP':
			type = ULTIPA.PropertyType.PROPERTY_TIMESTAMP
		elif type.upper() == 'PROPERTY_TEXT' or type.upper() == 'TEXT':
			type = ULTIPA.PropertyType.PROPERTY_TEXT
		elif type.upper() == 'PROPERTY_UUID' or type.upper() == 'UUID':
			type = ULTIPA.PropertyType.PROPERTY_UUID
		elif type.upper() == 'PROPERTY_ID' or type.upper() == 'ID':
			type = ULTIPA.PropertyType.PROPERTY_ID
		elif type.upper() == 'PROPERTY_FROM' or type.upper() == 'FROM':
			type = ULTIPA.PropertyType.PROPERTY_FROM
		elif type.upper() == 'PROPERTY_FROM_UUID' or type.upper() == 'FROM_UUID':
			type = ULTIPA.PropertyType.PROPERTY_FROM_UUID
		elif type.upper() == 'PROPERTY_TO' or type.upper() == 'TO':
			type = ULTIPA.PropertyType.PROPERTY_TO
		elif type.upper() == 'PROPERTY_TO_UUID' or type.upper() == 'TO_UUID':
			type = ULTIPA.PropertyType.PROPERTY_TO_UUID
		elif type.upper() == 'PROPERTY_IGNORE' or type.upper() == 'IGNORE':
			type = ULTIPA.PropertyType.PROPERTY_IGNORE

		return type

	@staticmethod
	def _formatTableHeader(headers):
		headerList = []
		headertype = {}
		if isinstance(headers, str):
			headers = json.loads(headers)
		if isinstance(headers, list) and isinstance(headers, list):
			for index, header in enumerate(headers):
				headdic = {}
				if isinstance(header, dict):
					type = FormatType.propertyType(header.get('type'))
					headdic.update({'property_name': header.get('name')})
					headdic.update({'property_type': type})
					headertype.update({header.get('name'): type})
					headerList.append(headdic)
				else:
					headdic.update({'property_name': header.name})
					headdic.update({'property_type': header.type})
					if header.name in headertype:
						headertype.update({header.name: header.type})
						headerList.append(headdic)
					else:
						raise ParameterException('Attribute names cannot be the same')
		else:
			raise ParameterException(err='The headers and rows must be list')

		return (headerList, headertype)

	@staticmethod
	def pathScheams(path):
		nodeSchemas = {}
		edgeSchemas = {}
		for header in path.node_table.schemas:
			schemaName = header.schema_name
			propetyList = []
			for index, property in enumerate(header.properties):
				subTypes = [Property._getStringByPropertyType(sbt) for sbt in property.sub_types]
				newproperty = ULTIPA.Property(property.property_name,subTypes=subTypes)
				newproperty.type = Property._getStringByPropertyType(property.property_type)
				propetyList.append(newproperty)
			nodeSchemas.update({schemaName: ULTIPA.Schema(schemaName,None,None,None, propetyList, None)})
		for header in path.edge_table.schemas:
			schemaName = header.schema_name
			propetyList = []
			for index, property in enumerate(header.properties):
				subTypes = [Property._getStringByPropertyType(sbt) for sbt in property.sub_types]
				newproperty = ULTIPA.Property(property.property_name,subTypes=subTypes)
				newproperty.type = Property._getStringByPropertyType(property.property_type)
				propetyList.append(newproperty)
			edgeSchemas.update({schemaName: ULTIPA.Schema(schemaName,None,None,None, propetyList, None)})
		return nodeSchemas, edgeSchemas

	@staticmethod
	def makeEntityNodeTable(schema: ULTIPA_REQUEST.Schema, entity_rows: List[ULTIPA.EntityRow],
							timeZoneOffset=None) -> ULTIPA.NodeEntityTable:

		nodetable = ULTIPA.NodeEntityTable([], [])
		nodetable.schemas.append({"schema_name": schema.name, "properties": []})

		for prop in schema.properties:
			if prop.isIdType() or prop.isIgnore():
				continue
			nodetable.schemas[0].get("properties").append(
				{"property_name": prop.name, "property_type": prop.type, "sub_types": prop.subTypes})
		for i, row in enumerate(entity_rows):
			append = True
			values = []
			FormatType.checkEntityNodeRow(i, schema, row)
			for proper in schema.properties:
				if proper.isIdType() or proper.isIgnore():
					continue
				if proper.name in row.values:
					value = row.values.get(proper.name)
					if isinstance(proper.type,str):
						ptype = proper.getPropertyTypeByString(proper.type)
					else:
						ptype = proper.type
					checkRet = TypeCheck.checkProperty(ptype, value)
					if isinstance(checkRet, bool):
						try:
							_seria = _Serialize(type=ptype, value=value, name=proper.name,
												timeZoneOffset=timeZoneOffset, subTypes=proper.subTypes)
							sdata = _seria.serialize()
							values.append(sdata)
						except Exception as e:
							if len(e.args) > 0 and "%s" in e.args[0]:
								e = e.args[0] % (proper.name, value)
							raise ParameterException(err=f"node row [{row._getIndex()}] error: {e}")

					else:
						raise ParameterException(
							err=checkRet % ("node", row._getIndex(), proper.name, value))
				else:
					raise ParameterException(
						err=f"node row [{row._getIndex()}] error: values doesn't contain property [{proper.name}].")
			if append:
				data = {"schema_name": schema.name, "values": values}
				if row.uuid:
					data.update({"uuid": row.uuid})
				if row.id:
					data.update({"id": row.id})

				nodetable.nodeRows.append(data)
		return nodetable

	@staticmethod
	def makeEntityEdgeTable(schema: ULTIPA_REQUEST.Schema, rows: List[ULTIPA.EntityRow],
							timeZoneOffset=None) -> ULTIPA.EdgeEntityTable:
		edgetable = ULTIPA.EdgeEntityTable([], [])
		edgetable.schemas.append({"schema_name": schema.name, "properties": []})
		for prop in schema.properties:
			if prop.isIdType() or prop.isIgnore():
				continue
			edgetable.schemas[0].get("properties").append({"property_name": prop.name, "property_type": prop.type})
		for i, row in enumerate(rows):
			append = True
			values = []
			FormatType.checkEntityEdgeRow(i, schema, row)
			for proper in schema.properties:
				if proper.isIdType() or proper.isIgnore():
					continue
				if proper.name in row.values:
					value = row.values.get(proper.name)
					if isinstance(proper.type, str):
						ptype = proper.getPropertyTypeByString(proper.type)
					else:
						ptype = proper.type
					checkRet = TypeCheck.checkProperty(ptype, value)
					if isinstance(checkRet, bool):
						try:
							_seria = _Serialize(type=ptype, value=value, name=proper.name, subTypes=proper.subTypes,
												timeZoneOffset=timeZoneOffset)
							sdata = _seria.serialize()
							values.append(sdata)
						except Exception as e:
							raise ParameterException(err=f"edge row [{row._getIndex()}] error:{e}.")
					else:
						raise ParameterException(
							err=checkRet % ("edge", row._getIndex(), proper.name, value))
				else:
					raise ParameterException(
						err=f"edge row [{row._getIndex()}] error: values doesn't contain property [{proper.name}].")

			if append:
				data = {"schema_name": schema.name, "values": values}
				if row.from_uuid:
					data.update({"from_uuid": row.from_uuid})
				if row.to_uuid:
					data.update({"to_uuid": row.to_uuid})
				if row.from_id:
					data.update({"from_id": row.from_id})
				if row.to_id:
					data.update({"to_id": row.to_id})
				if row.uuid:
					data.update({"uuid": row.uuid})
				edgetable.edgeRows.append(data)
		return edgetable

	@staticmethod
	def nodeAliases(_nodes, timeZone, timeZoneOffset) -> List[ULTIPA.NodeAlias]:
		nodeAliasesData = []
		if _nodes:
			for node in _nodes:
				nodesData = FormatType.nodeTable(node, timeZone, timeZoneOffset)
				nodeAliasData = ULTIPA.NodeAlias(alias=node.alias, nodes=nodesData)
				nodeAliasesData.append(nodeAliasData)
		return nodeAliasesData

	@staticmethod
	def graph(graph, timeZone, timeZoneOffset) -> ULTIPA.Graph:
		if graph:
			nodeTable = FormatType.nodeTable(graph, timeZone, timeZoneOffset)
			edgeTable = FormatType.edgeTable(graph, timeZone, timeZoneOffset)
			return ULTIPA.Graph(nodeTable,edgeTable)
		return ULTIPA.Graph(None,None)


	@staticmethod
	def graphAlias(graphs,timeZone, timeZoneOffset) -> List[ULTIPA.GraphAlias]:
		nodeAliasesData = []
		if graphs:
			for graph in graphs:
				graphData = FormatType.graph(graph.graph, timeZone, timeZoneOffset)
				nodeAliasData = ULTIPA.GraphAlias(alias=graph.alias,graph=graphData)
				nodeAliasesData.append(nodeAliasData)
		return nodeAliasesData




	@staticmethod
	def attrNodeEntityTable(data) -> List[ULTIPA.Node]:
		return FormatType.attrEntity(data, "node")

	@staticmethod
	def attrEdgeEntityTable(data) -> List[ULTIPA.Edge]:
		return FormatType.attrEntity(data, "edge")

	@staticmethod
	def edgeAliases(_edges, timeZone, timeZoneOffset) -> List[ULTIPA.EdgeAlias]:
		edgeAliasesData = []
		if _edges:
			for edge in _edges:
				edgesData = FormatType.edgeTable(edge, timeZone, timeZoneOffset)
				edgeAliasData = ULTIPA.EdgeAlias(alias=edge.alias, edges=edgesData)
				edgeAliasesData.append(edgeAliasData)
		return edgeAliasesData

	@staticmethod
	def values(_values) -> 'dict':
		value_dict = {}
		if _values:
			for value in _values:
				value_dict.update({value.key: value.value})
		return value_dict

	@classmethod
	def parseAttrListData(cls, datas) -> List[ULTIPA.PaserAttrListData]:
		PaserAttrRetList = []
		for data in datas:
			attrListRet = []
			pathListRet = []
			nodetRet = None
			edgeRet = None
			attrListData = AttrListData()
			attrListData.ParseFromString(data)
			if attrListData.type == ULTIPA.ResultType.RESULT_TYPE_ATTR:
				for attr in attrListData.attrs:
					attrListRet.append(cls.parseAttr(attr, None))

			if attrListData.type == ULTIPA.ResultType.RESULT_TYPE_PATH:
				pathListRet.extend(FormatType.attrPath(attrListData.paths, None, None))

			if attrListData.type == ULTIPA.ResultType.RESULT_TYPE_NODE:
				nodetRet = FormatType.attrNodeEntityTable(attrListData.nodes)

			if attrListData.type == ULTIPA.ResultType.RESULT_TYPE_EDGE:
				edgeRet = FormatType.attrEdgeEntityTable(attrListData.edges)

			PaserAttrRetList.append(
				ULTIPA.PaserAttrListData(attrListData.type, nodes=nodetRet, edges=edgeRet, paths=pathListRet,
										 attrs=attrListRet))
		return PaserAttrRetList

	@classmethod
	def parseAttr(cls, attr, aliasName) -> ULTIPA.Attr:
		valuesList = []
		attr_type = attr.value_type
		if attr_type == ULTIPA.PropertyType.PROPERTY_LIST:
			valuesList = FormatType.parseAttrListData(attr.values)
			return ULTIPA.Attr(alias=aliasName, values=valuesList,
									type=Property._getStringByPropertyType(attr_type))

		for att in attr.values:
			_seria = _Serialize(type=attr_type, value=att)
			sdata = _seria.unserialize()
			valuesList.append(sdata)
		return ULTIPA.Attr(alias=aliasName, values=valuesList,
								type=Property._getStringByPropertyType(attr_type))

	@classmethod
	def parseAttrAlias(cls, attrAlias) -> ULTIPA.Attr:
		return cls.parseAttr(attrAlias.attr, attrAlias.alias)

	@staticmethod
	def attrAlias1(_attrs) -> List[ULTIPA.Attr]:
		attr_list = []
		if _attrs:
			for attr in _attrs:
				attr_list.append(FormatType.parseAttrAlias(attr))
		return attr_list

	@staticmethod
	def attr(attr, timeZone, timeZoneOffset) -> ULTIPA.UltipaAttr:
		type = attr.value_type
		result = ULTIPA.UltipaAttr(type, None, has_attr_data=False, has_ultipa_data=False,
							 type_desc=Property._getStringByPropertyType(type))
		result.values = []
		for value in attr.values:
			if type == ULTIPA.PropertyType.PROPERTY_NULL:
				result.has_attr_data = True
				result.values.append(None)
				continue
			if type == ULTIPA.PropertyType.PROPERTY_LIST or type== ULTIPA.PropertyType.PROPERTY_SET:
				attrListData = AttrListData()
				attrListData.ParseFromString(value)
				result_type: any = attrListData.type
				if attrListData.is_null:
					result.values.append(None)
					continue

				if result_type == ULTIPA.ResultType.RESULT_TYPE_ATTR:
					attrs = []
					for attr in attrListData.attrs:
						att = FormatType.attr(attr, timeZone, timeZoneOffset)
						if not att.values:
							attrs.append(att.values)
						else:
							attrs.extend(att.values)
					if len(attrs) != 0:
						result.has_attr_data = True
					result.values.append(attrs)

				if result_type == ULTIPA.ResultType.RESULT_TYPE_PATH:
					result.has_ultipa_data = True
					result.values.append(ULTIPA.PaserAttrListData(attrListData.type,
																  paths=FormatType.attrPath(attrListData.paths,
																							timeZone, timeZoneOffset)))

				if result_type == ULTIPA.ResultType.RESULT_TYPE_NODE:
					result.has_ultipa_data = True
					result.values.append(ULTIPA.PaserAttrListData(attrListData.type,
																  nodes=FormatType.attrNodeEntityTable(
																	  attrListData.nodes)))

				if result_type == ULTIPA.ResultType.RESULT_TYPE_EDGE:
					result.has_ultipa_data = True
					result.values.append(ULTIPA.PaserAttrListData(attrListData.type,
																  edges=FormatType.attrEdgeEntityTable(
																	  attrListData.edges)))
			else:
				_seria = _Serialize(type=type, value=value, timeZone=timeZone, timeZoneOffset=timeZoneOffset)
				sdata = _seria.unserialize()
				result.has_attr_data = True
				result.values.append(sdata)
		return result

	@staticmethod
	def attrAlias(_attrs, timeZone, timeZoneOffset) -> List[ULTIPA.AttrNewAlias]:
		attr_list = []
		if _attrs:
			for attr in _attrs:
				attr_data = FormatType.attr(attr.attr, timeZone, timeZoneOffset)
				attr_list.append(ULTIPA.AttrNewAlias(alias=attr.alias, attr=attr_data))
		return attr_list

	@staticmethod
	def arrays(_arrays) -> List[ULTIPA.ArrayAlias]:
		arrays_list = []
		if _arrays:
			for attr in _arrays:
				li = []
				attr_type = attr.property_type
				for att in attr.elements:
					value_li = []
					for el in att.values:
						_seria = _Serialize(type=attr_type, value=el)
						sdata = _seria.unserialize()
						value_li.append(sdata)
					li.append(value_li)
				ret = ULTIPA.ArrayAlias(alias=attr.alias, elements=li)
				arrays_list.append(ret)
		return arrays_list

	@staticmethod
	def resultalias(_arrays) -> List[ULTIPA.ResultAlias]:
		value_dict = []
		if _arrays:
			for value in _arrays:
				value_dict.append(ULTIPA.ResultAlias(value.alias, ResultType.getTypeStr(value.result_type)))	
		return value_dict

	@staticmethod
	def explainPlan(_explainPlan) -> List[ULTIPA.ExplainPlan]:
		explainPlanRet = []
		if _explainPlan:
			for value in _explainPlan.plan_nodes:
				explainPlanRet.append(
					ULTIPA.ExplainPlan(value.alias, value.children_num, value.uql, value.infos))
		return explainPlanRet
	@staticmethod
	def export_edges(_edges, timeZone, timeZoneOffset) -> List:
		edgeTableData = FormatType.edgeTable(_edges, timeZone, timeZoneOffset)

		return edgeTableData

	@staticmethod
	def export_nodes(_nodes, timeZone, timeZoneOffset) -> List:
		nodeTableData = FormatType.nodeTable(_nodes, timeZone, timeZoneOffset)
		return nodeTableData


	@staticmethod
	def pathAlias(_paths, timeZone, timeZoneOffset) -> [ULTIPA.PathAlias]:
		pathData = []
		if _paths:
			for path in _paths:
				pathAlia = ULTIPA.PathAlias(path.alias)
				for npath in path.paths:
					nodeAliasesData = FormatType.nodeTable(npath, timeZone, timeZoneOffset)
					edgeAliasesData = FormatType.edgeTable(npath, timeZone, timeZoneOffset)
					nodeSchema, edgeSchema = FormatType.pathScheams(npath)
					newPath = ULTIPA.Path(nodes=nodeAliasesData, edges=edgeAliasesData, nodeSchemas=nodeSchema,
										  edgeSchemas=edgeSchema)
					pathAlia.paths.append(newPath)
				pathData.append(pathAlia)
		return pathData

	@staticmethod
	def attrPath(_paths, timeZone, timeZoneOffset) -> [ULTIPA.Path]:
		pathData = []
		if _paths:
			for path in _paths:
				nodeAliasesData = FormatType.nodeTable(path, timeZone, timeZoneOffset)
				edgeAliasesData = FormatType.edgeTable(path, timeZone, timeZoneOffset)
				nodeSchema, edgeSchema = FormatType.pathScheams(path)
				newPath = ULTIPA.Path(nodes=nodeAliasesData, edges=edgeAliasesData, nodeSchemas=nodeSchema,
									  edgeSchemas=edgeSchema)
				pathData.append(newPath)
		return pathData

	@staticmethod
	def Response(_res, host: str = None) -> ULTIPA_RESPONSE.Response:
		status = FormatType.status(_res.status, host)
		res = ULTIPA_RESPONSE.Response(status)
		return res

	@staticmethod
	def mergeUqlResponse(mergeRes: ULTIPA_RESPONSE.Response, res: ULTIPA_RESPONSE.Response):
		if not mergeRes.data:
			mergeRes = copy.deepcopy(res)
			return mergeRes
		if res.data.paths:
			mergeRes.data.paths = DataMerge.concat(mergeRes.data.paths, res.data.paths)
		if res.data.nodes:
			mergeRes.data.nodes = DataMerge.concat(mergeRes.data.nodes, res.data.nodes, 'nodes')
		if res.data.edges:
			mergeRes.data.edges = DataMerge.concat(mergeRes.data.edges, res.data.edges, 'edges')
		if res.data.attrs:
			mergeRes.data.attrs = DataMerge.concat(mergeRes.data.attrs, res.data.attrs, 'values')
		if res.data.tables:
			mergeRes.data.tables = DataMerge.concat(mergeRes.data.tables, res.data.tables, "table_name", ["headers"])
		if res.data.explainPlan:
			mergeRes.data.explainPlan = DataMerge.concat(mergeRes.data.explainPlan, res.data.explainPlan)
		if res.data.resultAlias:
			mergeRes.data.resultAlias = DataMerge.concat(mergeRes.data.resultAlias, res.data.resultAlias)
		if res.statistics:
			mergeRes.statistics = DataMerge.concat(mergeRes.statistics, res.statistics)

		return mergeRes

	@staticmethod
	def response(uql_response: ULTIPA_RESPONSE.Response, uqlReply, timeZone,
				 timeZoneOffset) -> ULTIPA_RESPONSE.Response:
		attrs = FormatType.attrAlias(uqlReply.attrs, timeZone, timeZoneOffset)		
		attrs_attrs = []
		attrs_map = {}

		def addAttributes(alias: str, type: ULTIPA.ResultType, values: any, dataMap: HasDataMap) -> None:
			if attrs_map.get(alias) is None:
				if dataMap.has_attr_data:
					type = ULTIPA.ResultType.RESULT_TYPE_ATTR
				attralias = ULTIPA.Attr(alias=alias, type=type, type_desc=ULTIPA.ResultType.getTypeStr(type),
											 values=[])
				attrs_map[alias] = attralias
				attrs_attrs.append(attralias)
			if type == ULTIPA.ResultType.RESULT_TYPE_ATTR and dataMap.only_attr_list:
				attrs_map[alias].values = values
			else:
				attrs_map[alias].values.append(values)

		for attrAlias in attrs:
			alias = attrAlias.alias
			has_attr_data = attrAlias.attr.has_attr_data
			has_ultipa_data = attrAlias.attr.has_ultipa_data
			only_attr_list = has_attr_data and not has_ultipa_data
			dataMap = HasDataMap(has_ultipa_data=has_ultipa_data, has_attr_data=has_attr_data,
								 only_attr_list=only_attr_list)
			if attrAlias.attr.type == ULTIPA.PropertyType.PROPERTY_LIST:
				if only_attr_list:
					addAttributes(alias, ULTIPA.ResultType.RESULT_TYPE_ATTR, attrAlias.attr.values, dataMap)
					continue

				for row in attrAlias.attr.values:
					if not row or not row.type:
						addAttributes(alias, ULTIPA.ResultType.RESULT_TYPE_ATTR, row, dataMap)
						continue

					if not isinstance(row, ULTIPA.PaserAttrListData):
						addAttributes(alias, ULTIPA.ResultType.RESULT_TYPE_ATTR, row, dataMap)
						continue

					if row.type == ULTIPA.ResultType.RESULT_TYPE_NODE:
						addAttributes(alias, row.type, row.nodes, dataMap)

					elif row.type == ULTIPA.ResultType.RESULT_TYPE_EDGE:
						addAttributes(alias, row.type, row.edges, dataMap)

					elif row.type == ULTIPA.ResultType.RESULT_TYPE_PATH:
						addAttributes(alias, row.type, row.paths, dataMap)

			else:
				addAttributes(alias, ULTIPA.ResultType.RESULT_TYPE_ATTR, attrAlias.attr.values, dataMap)

		status = FormatType.status(uqlReply.status)
		tables = FormatType.tables(uqlReply.tables, timeZone, timeZoneOffset)
		paths = FormatType.pathAlias(uqlReply.paths, timeZone, timeZoneOffset)
		graphs = FormatType.graphAlias(uqlReply.graphs, timeZone, timeZoneOffset)
		nodes = FormatType.nodeAliases(uqlReply.nodes, timeZone, timeZoneOffset)  # ForamtType.nodeAliases
		edges = FormatType.edgeAliases(uqlReply.edges, timeZone, timeZoneOffset)
		attrs = attrs_attrs
		statistics = FormatType.statistics(uqlReply.statistics, timeZone, timeZoneOffset)
		resultAlias = FormatType.resultalias(uqlReply.alias)
		explainPlan = FormatType.explainPlan(uqlReply.explain_plan)
		baseReply = ULTIPA.BaseUqlReply(paths=paths, nodes=nodes, edges=edges, tables=tables,
										attrs=attrs, resultAlias=resultAlias, explainPlan=explainPlan,graphs=graphs)
		uql_response.status = status
		uql_response.statistics = statistics
		uql_response.data = baseReply
		uql_response.aliases = resultAlias
		return uql_response

	@staticmethod
	def uqlResponse(_res, timeZone, timeZoneOffset) -> ULTIPA_RESPONSE.UltipaResponse:
		uql_response = ULTIPA_RESPONSE.Response()
		ultipa_response = ULTIPA_RESPONSE.UltipaResponse()
		for uqlReply in _res:
			status = FormatType.status(uqlReply.status)
			uql_response = FormatType.response(uql_response, uqlReply, timeZone, timeZoneOffset)
			ret = ULTIPA.UqlReply(dataBase=uql_response.data)

			if status.code != ULTIPA.Code.SUCCESS:
				ultipa_response.status = uql_response.status
				ultipa_response.req = uql_response.req
				return ultipa_response

			ultipa_response.items = ret._aliasMap
			ultipa_response.status = uql_response.status
			ultipa_response.req = uql_response.req
			ultipa_response.statistics = uql_response.statistics
			yield ultipa_response

	@staticmethod
	def uqlMergeResponse(_res, timeZone, timeZoneOffset) -> ULTIPA_RESPONSE.UltipaResponse:
		uql_response = ULTIPA_RESPONSE.Response()
		mergeRes = ULTIPA_RESPONSE.Response()
		ultipa_response = ULTIPA_RESPONSE.UltipaResponse()
		aliasvalue=[]
		for uqlReply in _res:
			status = FormatType.status(uqlReply.status)
			uql_response = FormatType.response(uql_response, uqlReply, timeZone, timeZoneOffset)
			if uql_response.aliases:
				aliasvalue.append(uql_response.aliases)
			if status.code != ULTIPA.Code.SUCCESS:
				ultipa_response.status = uql_response.status
				ultipa_response.req = uql_response.req
				return ultipa_response
			mergeRes = FormatType.mergeUqlResponse(mergeRes, uql_response)

		ret = ULTIPA.UqlReply(dataBase=mergeRes.data)
		ultipa_response.items = ret._aliasMap
		ultipa_response.explainPlan = ret.explainPlan
		ultipa_response.status = uql_response.status
		ultipa_response.req = uql_response.req
		if len(aliasvalue)>0:
			ultipa_response.aliases = aliasvalue[0]
		else:	
			ultipa_response.aliases = uql_response.aliases
		# ultipa_response.aliases = uql_response.aliases
		# ultipa_response.aliases = aliasvalue[0]
		ultipa_response.statistics = uql_response.statistics
		return ultipa_response

	@staticmethod
	def downloadResponse(_res) -> ULTIPA_RESPONSE.Response:
		for uqlReply in _res:
			status = FormatType.status(uqlReply.status)
			data = FormatType.status(uqlReply.chunk)
			total_time = uqlReply.total_time_cost
			engine_time = uqlReply.total_time_cost
			res = ULTIPA_RESPONSE.Response(status, total_time, engine_time, data)
			return res

	@staticmethod
	def exportResponse(_res, timeZone, timeZoneOffset) -> ULTIPA_RESPONSE.Response:
		nodedata = []
		edgedata = []
		res = ULTIPA_RESPONSE.Response()
		for uqlReply in _res:
			res.status = FormatType.status(uqlReply.status)
			if uqlReply.node_table:
				nodedata = FormatType.export_nodes(uqlReply, timeZone, timeZoneOffset)
			if uqlReply.edge_table:
				edgedata = FormatType.export_edges(uqlReply, timeZone, timeZoneOffset)
			if nodedata:
				uql = ULTIPA.ExportReply(data=nodedata)
				res.data = uql.data
			if edgedata:
				uql = ULTIPA.ExportReply(data=edgedata)
				res.data = uql.data
			break
		return res

	@staticmethod
	def graphPrivileges(graphPrivileges: '[object]'):
		new_graphPrivileges = []

		if graphPrivileges:
			for gp in graphPrivileges:
				graphDcit = {}
				graphDcit.update({'name': list(gp.keys())[0]})
				graphDcit.update({'values': list(gp.values())[0]})
				new_graphPrivileges.append(graphDcit)

		return new_graphPrivileges




from typing import List, Dict
from ultipa.printer.prettyPrint import PrettyPrint, PrettyTable
from ultipa.structs.GraphSet import GraphSet
from ultipa.structs.Property import Property
from ultipa.structs.Schema import Schema
from ultipa.structs.DBType import DBType
from ultipa.structs.User import User
from ultipa.structs.Policy import Policy
from ultipa.structs.Privilege import Privilege
from ultipa.structs.Index import Index
from ultipa.structs.Stats import Stats
from ultipa.structs.Top import Top
from ultipa.structs.Algo import Algo
from ultipa.types.types import BaseModel, Node, Edge, DataItem, Status, \
	UltipaStatistics, ReturnReq, BaseUqlReply, ResultType, ResultAlias, ExplainPlan
from ultipa.utils.convert import Any
from typing import Callable



class PropertyTable(BaseModel):
	name: str
	data: List[Property]

	def __init__(self, name, data):
		self.name = name
		self.data = data



class Response(BaseModel):
	def __init__(self, status: Status = None, data: BaseUqlReply = None,
				 req: ReturnReq = None, statistics: UltipaStatistics = None, aliases: List[ResultAlias] = None):
		self.status = status
		self.data = data
		self.statistics = statistics
		self.req = req
		self.aliases = aliases

	def Print(self):
		pretty = PrettyPrint()
		pretty.prettyStatus(self.status)
		if self.status.code != 0:
			return
		if self.statistics:
			pretty.prettyStatistics(self.statistics)

		if isinstance(self.data, list):
			dict_list = []
			for i in self.data:
				if isinstance(i,PropertyTable):
					dict_list.append(i.toDict())
				else:
					dict_list.append(i.__dict__)
			pretty.prettyDataList(dict_list)

		if isinstance(self.data, Any):
			dict_list = self.data.__dict__
			pretty.prettyData(dict_list)


class UltipaResponse(BaseModel):
	def __init__(self, status: Status = None, items: Dict = None,
				 req: ReturnReq = None, statistics: UltipaStatistics = None, aliases: List[ResultAlias] = None,
				 explainPlan: List[ExplainPlan] = None):
		self.status = status
		self.items = items
		self.aliases = aliases
		self.req = req
		self.statistics = statistics
		self.explainPlan = explainPlan

	def alias(self, alias: str) -> DataItem:
		if self.items == None:
			return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))
		if self.items.get(alias):
			return self.items.get(alias)
		return DataItem(alias, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))

	def get(self, index: int) -> DataItem:
		if len(self.aliases) - 1 >= index:
			data = self.items.get(self.aliases[index].alias)
			if data:
				return data
			if self.aliases[index].result_type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
				return DataItem(None, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))
			else:
				return data
		return DataItem(None, None, ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET))

	def getExplainPlan(self):
		return self.explainPlan

	def Print(self):
		pretty = PrettyPrint()
		pretty.prettyStatus(self.status)
		if self.status.code != 0:
			return
		pretty.prettyStatistics(self.statistics)
		explains = []
		for key in self.items:
			dataItem = self.items.get(key)
			if dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
				pretty.prettyNode(dataItem)

			elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
				pretty.prettyEdge(dataItem)

			elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
				pretty.prettyTable(dataItem)

			elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH):
				pretty.prettyPath(dataItem)

			elif dataItem.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR):
				pretty.prettyAttr(dataItem)


		if self.explainPlan:
			for explan in self.explainPlan:
				explains.append(explan)
			if explains:
				pretty.prettyTree(explains)

class UQLResponseStream:
    def __init__(self):
        self._listeners = {}
        self._default_handlers = {
            "start": self._default_start_handler,
            "end": self._default_end_handler
        }

    def on(self, event: str, handler: Callable):
        if event not in self._listeners:
            self._listeners[event] = []
        self._listeners[event].append(handler)

    def emit(self, event: str, *args, **kwargs):
        if event in self._listeners:
            for handler in self._listeners[event]:
                handler(*args, **kwargs)
        elif event in self._default_handlers:
            self._default_handlers[event](*args, **kwargs)

    def _default_start_handler(self, request_config):
        print("Stream started with request:", request_config)

    def _default_end_handler(self, request_config):
        print("Stream ended with request:", request_config)


class ResponseCommon(Response):
	data: None



# class User(BaseModel):
# 	username: str
# 	create: str
# 	last_login_time: str
# 	graphPrivileges: dict
# 	systemPrivileges: List[str]
# 	policies: List[str]


# class Index(BaseModel):
# 	name: str
# 	properties: str
# 	schema: str
# 	status: str
# 	size:str
# 	dbType:DBType


# class IndexTable:
# 	name: str
# 	data: List[Index]


# class Stat(BaseModel):
# 	cpuUsage: str
# 	memUsage: str
# 	company: str
# 	cpuCores: str
# 	expiredDate: str
# 	serverType: str
# 	version: str




# class Privilege(BaseModel):
# 	graphPrivileges: List[str]
# 	systemPrivileges: List[str]


# class Policy(BaseModel):
# 	name: str
# 	policies: List[str]
# 	graphPrivileges: dict
# 	systemPrivileges: List[str]


# class Top(BaseModel):
# 	process_id: str
# 	process_uql: str
# 	duration: str


class Return_Type(BaseModel):
	is_realtime: bool
	is_visualization: bool
	is_wirte_back: bool


class Task_info(BaseModel):
	task_id: int
	algo_name: str
	start_time: int
	writing_start_time: int
	end_time: int
	time_cost: int
	TASK_STATUS: int
	status_code: str
	engine_cost: int
	return_type: Return_Type


class Task(BaseModel):
	param: dict
	task_info: Task_info
	result: dict


class SearchKhop(BaseModel):
	values: dict
	nodes: List[Node]


class Path(BaseModel):
	nodes: List[Node]
	edges: List[Edge]


class SearchPath(BaseModel):
	paths: List[Path]


class NodeSpread(SearchPath):
	pass


class AutoNet(SearchPath):
	pass


class AlgoResultOpt(BaseModel):
	can_realtime: bool
	can_visualization: bool
	can_write_back: bool

class Exta(BaseModel):
	author:str
	detail:str
	name:str
	version:str

# class Algo(BaseModel):
# 	param: dict
# 	name: str
# 	result_opt: AlgoResultOpt

class ResponseGraph(Response):
	data: GraphSet

class ResponseListGraph(Response):
	data: List[GraphSet]
	

class ResponeListExta(Response):
	data: List[Exta]	

class ResponseSchema(Response):
	items:Schema

class ResponseListSchema(UltipaResponse):
	items:List[Schema]	


class ResponseListIndex(UltipaResponse):
	data: List[Index]


# class ResponseListFulltextIndex(ResponseListIndex):
# 	pass


class ResponseSearchEdge(UltipaResponse):
	items: List[DataItem]


class ResponseSearchNode(UltipaResponse):
	items: List[DataItem]


class ResponseBulk:
	uuids: List[int]
	errorItem: Dict


class ResponseInsertNode(Response):
	data: List[Node]

class ResponseInsertEdge(Response):
	data: List[Edge]	

class ResponseDeleteNode(Response):
	data: List[Node]

class ResponseDeleteEdge(Response):
	data: List[Edge]		



class ResponseBatchAutoInsert(Response):
	data: Dict[str, ResponseBulk]


class InsertResponse(Response):
	data: ResponseBulk


class ResponseListPolicy(Response):
	data: List[Policy]


class ResponsePolicy(Response):
	data: Policy


class ResponsePrivilege(Response):
	data: List[Privilege]


class ResponseListProperty(UltipaResponse):
	items: List[Property]


class ResponseProperty(Response):
	data: List[Property]


class ResponseListTop(Response):
	data: List[Top]


class ResponseListTask(Response):
	data: List[Task]


class ResponseUser(Response):
	data: User


class ResponseListUser(Response):
	data: List[User]


class ResponseListAlgo(Response):
	data: List[Algo]


class Cluster:
	host: str
	status: bool
	cpuUsage: str
	memUsage: str
	isLeader: bool
	isFollowerReadable: bool
	isAlgoExecutable: bool
	isUnset: bool


class ClusterInfo(Response):
	data: List[Cluster]


class ResponseStat(Response):
	data: Stats


class ResponseExport(Response):
	data: List[Node]





from ultipa import DBType
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.structs.Property import Property
from ultipa.structs.Index import Index
from ultipa.utils import UQLMAKER, CommandList
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs.DBType import DBType
from ultipa.utils.convert import convertToIndex
from ultipa.utils.convert import convertToFullText

class IndexExtra(BaseExtra):

	'''
	Processing class that defines settings for index related operations.
	'''

	def showIndex(self,
				  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all indice.

		Args:
			requestConfig: An object of RequestConfig class
			
		Returns:
			ResponseListIndex
		'''
		# if dbType != None:
		# 	command = dbType == DBType.DBNODE and CommandList.showNodeIndex or CommandList.showEdgeIndex
		# else:
		command = CommandList.showIndex
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		indexdata=convertToIndex(res=res,all=True)
		res.data=indexdata
		return res
	
	def showNodeIndex(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all Node indice.

		Args:
			requestConfig: An object of RequestConfig class
			
		Returns:
			ResponseListIndex
		'''
		# res=self.showIndex(dbType=DBType.DBNODE,requestConfig=requestConfig)
		command = CommandList.showNodeIndex
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		indexdata=convertToIndex(res=res,all=False,dbtype=DBType.DBNODE)
		res.data=indexdata
		return res
	
	def showEdgeIndex(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all Edge indice.

		Args:
			requestConfig: An object of RequestConfig class
			
		Returns:
			ResponseListIndex
		'''
		# res=self.showIndex(dbType=DBType.DBEDGE,requestConfig=requestConfig)
		command = CommandList.showEdgeIndex
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		indexdata=convertToIndex(res=res,all=False,dbtype=DBType.DBEDGE)
		res.data=indexdata
		return res
	

	def showFulltext(self,
					 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all full-text indice.

		Args:
			requestConfig: An object of RequestConfig class
		Returns:
		
			ResponseListIndex
		'''
		# if dbType != None:
		# 	command = dbType == DBType.DBNODE and CommandList.showNodeFulltext or CommandList.showEdgeFulltext
		# else:
		command = CommandList.showFulltext
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		fulltextdata=convertToFullText(res=res)
		res.data=fulltextdata
		return res
	
	def showNodeFulltext(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all full-text Node indice.

		Args:

			requestConfig: An object of RequestConfig class

		Returns:
		
			ResponseListIndex
		'''
		# res=self.showFulltext(dbType=DBType.DBNODE,requestConfig=requestConfig)
		command = CommandList.showNodeFulltext
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		fulltextdata=convertToFullText(res=res,all=False,dbtype=DBType.DBNODE)
		res.data=fulltextdata
		return res
	
	def showEdgeFulltext(self,requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListIndex:
		'''
		Show all full-text Edge indice.

		Args:

			requestConfig: An object of RequestConfig class

		Returns:
		
			ResponseListIndex
		'''
		# res=self.showFulltext(dbType=DBType.DBEDGE,requestConfig=requestConfig)
		command = CommandList.showEdgeFulltext
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=False)
		fulltextdata=convertToFullText(res=res,all=False,dbtype=DBType.DBEDGE)
		res.data=fulltextdata
		return res
	
	

	def createIndex(self, dbType:DBType,propertyName:str,schemaName:str=None,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create an index.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			property: An object of Property class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.createNodeIndex or CommandList.createEdgeIndex
		# commandP = request.toString
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" % (propertyName)	
		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def createFulltext(self, dbType:DBType,schemaName:str,propertyName:str,fulltextName:str,
					   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a full-text index.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			property: An object of Property class

			fulltextName: Name of the fulltext index

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		command = dbType == DBType.DBNODE and CommandList.createNodeFulltext or CommandList.createEdgeFulltext
		command1="@`%s`.`%s`" % (schemaName, propertyName)
		commandP = [command1,fulltextName]
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropIndex(self, dbType:DBType,propertyName:str,schemaName:str=None,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop an index.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			propertyName: The name of property

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''
		command = dbType == DBType.DBNODE and CommandList.dropNodeIndex or CommandList.dropEdgeIndex
		# commandP = request.toString
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" % (propertyName)

		uqlMaker = UQLMAKER(command=command, commandP=commandP, commonParams=requestConfig)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropFulltext(self,dbType:DBType,fulltextName:str,
					 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop a full-text index.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			fulltextName: Name of the fulltext index

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.dropNodeFulltext or CommandList.dropEdgeFulltext
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(fulltextName)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:19
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Edge.py
from typing import Dict

from ultipa.structs.BaseModel import BaseModel


class Edge(BaseModel):
	'''
	    Data class for edge.
	'''
	_index = None

	def __init__(self, values: Dict, from_id: str = None, from_uuid: int = None, to_id: str = None, to_uuid: int = None,
				 schema: str = None,
				 uuid: int = None, **kwargs):
		if schema is None:
			if kwargs.get("schema_name") is not None:
				self.schema = kwargs.get("schema_name")
			else:
				self.schema = None
		else:
			self.schema = schema

		self.from_id = from_id
		self.from_uuid = from_uuid
		self.to_id = to_id
		self.to_uuid = to_uuid
		self.values = values
		self.uuid = uuid

	def getUUID(self):
		return self.uuid

	def getFrom(self):
		return self.from_id

	def getTo(self):
		return self.to_id

	def getFromUUID(self):
		return self.from_uuid

	def getToUUID(self):
		return self.to_uuid

	def getValues(self):
		return self.values

	def getSchema(self):
		return self.schema

	def get(self, propertyName: str):
		return self.values.get(propertyName)

	def set(self, propertyName: str, value):
		self.values.update({propertyName: value})

	def _getIndex(self):
		return self._index